"""
  Gradient boosting is a classic sequential ensemble method. At each iteration,
  the learning target of a new base estimator is to fit the pseudo residuals
  computed based on the ground truth and the output from base estimators
  fitted before, using ordinary least square.
"""


import abc
import torch
import logging
import warnings
import torch.nn as nn
import torch.nn.functional as F

from ._base import BaseModule, BaseClassifier, BaseRegressor
from ._base import torchensemble_model_doc
from .utils import io
from .utils import set_module
from .utils import operator as op
from .utils.logging import get_tb_logger


__all__ = ["GradientBoostingClassifier", "GradientBoostingRegressor"]


__model_doc = """
    Parameters
    ----------
    estimator : torch.nn.Module
        The class or object of your base estimator.

        - If :obj:`class`, it should inherit from :mod:`torch.nn.Module`.
        - If :obj:`object`, it should be instantiated from a class inherited
          from :mod:`torch.nn.Module`.
    n_estimators : int
        The number of base estimators in the ensemble.
    estimator_args : dict, default=None
        The dictionary of hyper-parameters used to instantiate base
        estimators. This parameter will have no effect if ``estimator`` is a
        base estimator object after instantiation.
    shrinkage_rate : float, default=1
        The shrinkage rate used in gradient boosting.
    cuda : bool, default=True

        - If ``True``, use GPU to train and evaluate the ensemble.
        - If ``False``, use CPU to train and evaluate the ensemble.

    Attributes
    ----------
    estimators_ : torch.nn.ModuleList
        An internal container that stores all fitted base estimators.
"""


__fit_doc = """
    Parameters
    ----------
    train_loader : torch.utils.data.DataLoader
        A data loader that contains the training data.
    epochs : int, default=100
        The number of training epochs per base estimator.
    use_reduction_sum : bool, default=True
        Whether to set ``reduction="sum"`` for the internal mean squared
        error used to fit each base estimator.
    log_interval : int, default=100
        The number of batches to wait before logging the training status.
    test_loader : torch.utils.data.DataLoader, default=None
        A data loader that contains the evaluating data.

        - If ``None``, no validation is conducted after each base
          estimator being trained.
        - If not ``None``, the ensemble will be evaluated on this
          dataloader after each base estimator being trained.
    early_stopping_rounds : int, default=2
        Specify the number of tolerant rounds for early stopping. When the
        validation performance of the ensemble does not improve after
        adding the base estimator fitted in current iteration, the internal
        counter on early stopping will increase by one. When the value of
        the internal counter reaches ``early_stopping_rounds``, the
        training stage will terminate instantly.
    save_model : bool, default=True
        Specify whether to save the model parameters.

        - If test_loader is ``None``, the ensemble containing
          ``n_estimators`` base estimators will be saved.
        - If test_loader is not ``None``, the ensemble with the best
          validation performance will be saved.
    save_dir : string, default=None
        Specify where to save the model parameters.

        - If ``None``, the model will be saved in the current directory.
        - If not ``None``, the model will be saved in the specified
          directory: ``save_dir``.
"""


def _gradient_boosting_model_doc(header, item="model"):
    """
    Decorator on obtaining documentation for different gradient boosting
    models.
    """

    def get_doc(item):
        """Return the selected item"""
        __doc = {"model": __model_doc, "fit": __fit_doc}
        return __doc[item]

    def adddoc(cls):
        doc = [header + "\n\n"]
        doc.extend(get_doc(item))
        cls.__doc__ = "".join(doc)
        return cls

    return adddoc


class _BaseGradientBoosting(BaseModule):
    def __init__(
        self,
        estimator,
        n_estimators,
        estimator_args=None,
        shrinkage_rate=1.0,
        cuda=True,
    ):
        super(BaseModule, self).__init__()
        self.base_estimator_ = estimator
        self.n_estimators = n_estimators
        self.estimator_args = estimator_args

        if estimator_args and not isinstance(estimator, type):
            msg = (
                "The input `estimator_args` will have no effect since"
                " `estimator` is already an object after instantiation."
            )
            warnings.warn(msg, RuntimeWarning)

        self.shrinkage_rate = shrinkage_rate
        self.device = torch.device("cuda" if cuda else "cpu")
        self.logger = logging.getLogger()
        self.tb_logger = get_tb_logger()

        self.estimators_ = nn.ModuleList()
        self.use_scheduler_ = False

    def _validate_parameters(
        self, epochs, log_interval, early_stopping_rounds
    ):
        """Validate hyper-parameters on training the ensemble."""

        if not epochs > 0:
            msg = (
                "The number of training epochs = {} should be strictly"
                " positive."
            )
            self.logger.error(msg.format(epochs))
            raise ValueError(msg.format(epochs))

        if not log_interval > 0:
            msg = (
                "The number of batches to wait before printting the"
                " training status should be strictly positive, but got {}"
                " instead."
            )
            self.logger.error(msg.format(log_interval))
            raise ValueError(msg.format(log_interval))

        if not early_stopping_rounds >= 1:
            msg = (
                "The number of tolerant rounds before triggering the"
                " early stopping should at least be 1, but got {} instead."
            )
            self.logger.error(msg.format(early_stopping_rounds))
            raise ValueError(msg.format(early_stopping_rounds))

        if not 0 < self.shrinkage_rate <= 1:
            msg = (
                "The shrinkage rate should be in the range (0, 1], but got"
                " {} instead."
            )
            self.logger.error(msg.format(self.shrinkage_rate))
            raise ValueError(msg.format(self.shrinkage_rate))

    @abc.abstractmethod
    def _handle_early_stopping(self, test_loader, est_idx):
        """Decide whether to trigger the internal counter on early stopping."""

    def _staged_forward(self, est_idx, *x):
        """
        Return the accumulated outputs from the first `est_idx+1` base
        estimators.
        """
        if est_idx >= self.n_estimators:
            msg = (
                "est_idx = {} should be an integer smaller than the"
                " number of base estimators = {}."
            )
            self.logger.error(msg.format(est_idx, self.n_estimators))
            raise ValueError(msg.format(est_idx, self.n_estimators))

        outputs = [
            estimator(*x) for estimator in self.estimators_[: est_idx + 1]
        ]
        out = op.sum_with_multiplicative(outputs, self.shrinkage_rate)

        return out

    def fit(
        self,
        train_loader,
        epochs=100,
        use_reduction_sum=True,
        log_interval=100,
        test_loader=None,
        early_stopping_rounds=2,
        save_model=True,
        save_dir=None,
    ):

        # Instantiate base estimators and set attributes
        for _ in range(self.n_estimators):
            self.estimators_.append(self._make_estimator())
        self._validate_parameters(epochs, log_interval, early_stopping_rounds)
        self.n_outputs = self._decide_n_outputs(train_loader)

        # Utils
        criterion = (
            nn.MSELoss(reduction="sum") if use_reduction_sum else nn.MSELoss()
        )
        n_counter = 0  # a counter on early stopping

        for est_idx, estimator in enumerate(self.estimators_):

            # Initialize a optimizer and scheduler for each base estimator to
            # avoid unexpected dependencies.
            learner_optimizer = set_module.set_optimizer(
                estimator, self.optimizer_name, **self.optimizer_args
            )

            if self.use_scheduler_:
                learner_scheduler = set_module.set_scheduler(
                    learner_optimizer,
                    self.scheduler_name,
                    **self.scheduler_args  # noqa: E501
                )

            # Training loop
            estimator.train()
            total_iters = 0
            for epoch in range(epochs):
                for batch_idx, elem in enumerate(train_loader):

                    data, target = io.split_data_target(elem, self.device)

                    # Compute the learning target of the current estimator
                    residual = self._pseudo_residual(est_idx, target, *data)

                    output = estimator(*data)
                    loss = criterion(output, residual)

                    learner_optimizer.zero_grad()
                    loss.backward()
                    learner_optimizer.step()

                    # Print training status
                    if batch_idx % log_interval == 0:
                        msg = (
                            "Estimator: {:03d} | Epoch: {:03d} | Batch:"
                            " {:03d} | RegLoss: {:.5f}"
                        )
                        self.logger.info(
                            msg.format(est_idx, epoch, batch_idx, loss)
                        )
                        if self.tb_logger:
                            self.tb_logger.add_scalar(
                                "gradient_boosting/Est_{}/Train_Loss".format(
                                    est_idx
                                ),
                                loss,
                                total_iters,
                            )
                    total_iters += 1

                if self.use_scheduler_:
                    learner_scheduler.step()

            # Validation
            if test_loader:
                flag = self._handle_early_stopping(test_loader, est_idx)

                if flag:
                    n_counter += 1
                    msg = "Early stopping counter: {} out of {}"
                    self.logger.info(
                        msg.format(n_counter, early_stopping_rounds)
                    )

                    if n_counter == early_stopping_rounds:
                        msg = "Handling early stopping..."
                        self.logger.info(msg)

                        # Early stopping
                        offset = est_idx - n_counter
                        self.estimators_ = self.estimators_[: offset + 1]
                        self.n_estimators = len(self.estimators_)
                        break
                else:
                    # Reset the counter if the performance improves
                    n_counter = 0

        # Post-processing
        msg = "The optimal number of base estimators: {}"
        self.logger.info(msg.format(len(self.estimators_)))
        if save_model:
            io.save(self, save_dir, self.logger)


@_gradient_boosting_model_doc(
    """Implementation on the GradientBoostingClassifier.""", "model"
)
class GradientBoostingClassifier(_BaseGradientBoosting, BaseClassifier):
    def _pseudo_residual(self, est_idx, y, *x):
        """Compute pseudo residuals in classification."""
        output = torch.zeros(y.size(0), self.n_outputs).to(self.device)

        # Before fitting the first estimator, we simply assume that GBM
        # outputs 0 for any input (i.e., a null output).
        if est_idx > 0:
            results = [
                estimator(*x) for estimator in self.estimators_[:est_idx]
            ]
            output += op.sum_with_multiplicative(results, self.shrinkage_rate)
        pseudo_residual = op.pseudo_residual_classification(
            y, output, self.n_outputs
        )

        return pseudo_residual

    def _handle_early_stopping(self, test_loader, est_idx):
        # Compute the validation accuracy of base estimators fitted so far
        self.eval()
        correct = 0
        total = 0
        flag = False
        with torch.no_grad():
            for _, elem in enumerate(test_loader):
                data, target = io.split_data_target(elem, self.device)
                output = F.softmax(self._staged_forward(est_idx, *data), dim=1)
                _, predicted = torch.max(output.data, 1)
                correct += (predicted == target).sum().item()
                total += target.size(0)
        acc = 100 * correct / total

        if est_idx == 0:
            self.best_acc = acc
        else:
            if acc > self.best_acc:
                self.best_acc = acc
            else:
                flag = True

        msg = "Validation Acc: {:.3f} % | Historical Best: {:.3f} %"
        self.logger.info(msg.format(acc, self.best_acc))
        if self.tb_logger:
            self.tb_logger.add_scalar(
                "gradient_boosting/Validation_Acc", acc, est_idx
            )

        return flag

    @torchensemble_model_doc(
        """Set the attributes on optimizer for GradientBoostingClassifier.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for GradientBoostingClassifier.""",
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @_gradient_boosting_model_doc(
        """Implementation on the training stage of GradientBoostingClassifier.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        use_reduction_sum=True,
        log_interval=100,
        test_loader=None,
        early_stopping_rounds=2,
        save_model=True,
        save_dir=None,
    ):
        self._criterion = nn.CrossEntropyLoss()
        super().fit(
            train_loader=train_loader,
            epochs=epochs,
            use_reduction_sum=use_reduction_sum,
            log_interval=log_interval,
            test_loader=test_loader,
            early_stopping_rounds=early_stopping_rounds,
            save_model=save_model,
            save_dir=save_dir,
        )

    @torchensemble_model_doc(
        """Implementation on the data forwarding in GradientBoostingClassifier.""",  # noqa: E501
        "classifier_forward",
    )
    def forward(self, *x):
        output = [estimator(*x) for estimator in self.estimators_]
        output = op.sum_with_multiplicative(output, self.shrinkage_rate)
        proba = F.softmax(output, dim=1)

        return proba

    @torchensemble_model_doc(item="classifier_evaluate")
    def evaluate(self, test_loader, return_loss=False):
        return super().evaluate(test_loader, return_loss)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)


@_gradient_boosting_model_doc(
    """Implementation on the GradientBoostingRegressor.""", "model"
)
class GradientBoostingRegressor(_BaseGradientBoosting, BaseRegressor):
    def _pseudo_residual(self, est_idx, y, *x):
        """Compute pseudo residuals in regression."""
        output = torch.zeros_like(y).to(self.device)

        if est_idx > 0:
            results = [
                estimator(*x) for estimator in self.estimators_[:est_idx]
            ]
            output = op.sum_with_multiplicative(results, self.shrinkage_rate)
        pseudo_residual = op.pseudo_residual_regression(y, output)

        return pseudo_residual

    def _handle_early_stopping(self, test_loader, est_idx):
        # Compute the validation MSE of base estimators fitted so far
        self.eval()
        mse = 0.0
        flag = False
        criterion = nn.MSELoss()
        with torch.no_grad():
            for _, elem in enumerate(test_loader):
                data, target = io.split_data_target(elem, self.device)
                output = self._staged_forward(est_idx, *data)
                mse += criterion(output, target)
        mse /= len(test_loader)

        if est_idx == 0:
            self.best_mse = mse
        else:
            assert hasattr(self, "best_mse")
            if mse < self.best_mse:
                self.best_mse = mse
            else:
                flag = True

        msg = "Validation MSE: {:.5f} | Historical Best: {:.5f}"
        self.logger.info(msg.format(mse, self.best_mse))
        if self.tb_logger:
            self.tb_logger.add_scalar(
                "gradient_boosting/Validation_MSE", mse, est_idx
            )

        return flag

    @torchensemble_model_doc(
        """Set the attributes on optimizer for GradientBoostingRegressor.""",
        "set_optimizer",
    )
    def set_optimizer(self, optimizer_name, **kwargs):
        super().set_optimizer(optimizer_name, **kwargs)

    @torchensemble_model_doc(
        """Set the attributes on scheduler for GradientBoostingRegressor.""",
        "set_scheduler",
    )
    def set_scheduler(self, scheduler_name, **kwargs):
        super().set_scheduler(scheduler_name, **kwargs)

    @_gradient_boosting_model_doc(
        """Implementation on the training stage of GradientBoostingRegressor.""",  # noqa: E501
        "fit",
    )
    def fit(
        self,
        train_loader,
        epochs=100,
        use_reduction_sum=True,
        log_interval=100,
        test_loader=None,
        early_stopping_rounds=2,
        save_model=True,
        save_dir=None,
    ):
        self._criterion = nn.MSELoss()
        super().fit(
            train_loader=train_loader,
            epochs=epochs,
            use_reduction_sum=use_reduction_sum,
            log_interval=log_interval,
            test_loader=test_loader,
            early_stopping_rounds=early_stopping_rounds,
            save_model=save_model,
            save_dir=save_dir,
        )

    @torchensemble_model_doc(
        """Implementation on the data forwarding in GradientBoostingRegressor.""",  # noqa: E501
        "regressor_forward",
    )
    def forward(self, *x):
        outputs = [estimator(*x) for estimator in self.estimators_]
        pred = op.sum_with_multiplicative(outputs, self.shrinkage_rate)

        return pred

    @torchensemble_model_doc(item="regressor_evaluate")
    def evaluate(self, test_loader):
        return super().evaluate(test_loader)

    @torchensemble_model_doc(item="predict")
    def predict(self, *x):
        return super().predict(*x)



"""
This is a random ensemble hybrid of DDPG/TD3, roughly based on https://arxiv.org/pdf/1907.04543.pdf.
"""

from mrl.algorithms.continuous_off_policy import *

class RandomEnsembleDPG(OffPolicyActorCritic):
  def optimize_from_batch(self, states, actions, rewards, next_states,
                        gammas):
    config = self.config

    a_next_max = self.actor_target(next_states)
    noise = torch.randn_like(a_next_max) * (
        config.td3_noise * self.action_scale)
    noise = noise.clamp(-config.td3_noise_clip * self.action_scale,
                        config.td3_noise_clip * self.action_scale)
    a_next_max = (a_next_max + noise).clamp(-self.action_scale,
                                            self.action_scale)

    qs = []
    for critic in self.critics:
      qs.append(critic(next_states, a_next_max))

    qs = torch.cat(qs, dim=-1) # batch x num_qs

    sample = torch.distributions.dirichlet.Dirichlet(torch.ones(1, qs.size(-1))).sample().to(self.config.device)

    target = (rewards + gammas * (qs * sample).sum(-1, keepdim=True))
    target = torch.clamp(target, *self.config.clip_target_range).detach()

    if hasattr(self, 'logger') and self.config.opt_steps % 100 == 0:
      self.logger.add_histogram('Optimize/Target_q', target)

    qs = []
    for critic in self.critics:
      qs.append(critic(states, actions))
    qs = (torch.cat(qs, dim=-1) * sample).sum(-1, keepdim=True)
    
    critic_loss = F.mse_loss(qs, target)

    self.critic_opt.zero_grad()
    critic_loss.backward()
    self.critic_opt.step()

    if config.opt_steps % config.td3_delay == 0:
      a = self.actor(states)
      qs = []
      for critic in self.critics:
        qs.append(critic(states, a))
      qs = torch.cat(qs, dim=-1)
      
      actor_loss = -qs.mean()

      self.actor_opt.zero_grad()
    actor_loss.backward()
    self.actor_opt.step()


import torch as t
from numpy.random import Generator
from rich.progress import track
from torch.optim import Adam
from torch.utils.data import DataLoader, Dataset

from rrt_ml.algorithms.base import *
from rrt_ml.utilities.datasets import *
from rrt_ml.utilities.formulas import *
from rrt_ml.utilities.infos import *
from rrt_ml.utilities.maps import *
from rrt_ml.utilities.models import *
from rrt_ml.utilities.stats import *


class SL(Algorithm):
    """
    Algorithm to learn to sample states for motion planning.
    """

    def __init__(self, cfg):
        """
        Initialize.
        """

        super(SL, self).__init__(cfg)
        self.dataset_train = None  # type: None | Dataset
        self.dataset_val = None  # type: None | Dataset
        self.dataloader_train = None  # type: None | DataLoader
        self.dataloader_val = None  # type: None | DataLoader

        self._setup()

    def train(self):
        """
        Train CVAE.
        """

        # Training loop
        for epoch_n in track(range(self.epoch + 1, self.cfg.sl.train.n_epochs), "Training CVAE..."):

            # Initialize list of stats
            t_l, v_l = [], []
            t_kl_l, v_kl_l = [], []
            t_r_l, v_r_l = [], []

            # Train
            for batch_n, sample in enumerate(iter(self.dataloader_train)):
                l, kl, r = self._set_train_on_batch(sample)
                t_l.append(l)
                t_kl_l.append(kl)
                t_r_l.append(r)

            # Validate
            for batch_n, sample in enumerate(iter(self.dataloader_val)):
                l, kl, r = self._set_validate_on_batch(sample)
                v_l.append(l)
                v_kl_l.append(kl)
                v_r_l.append(r)

            # Get all epoch info and set validation loss attributes
            epoch_info = SLEpochInfo.new(epoch_n, t_l, v_l, t_kl_l, v_kl_l, t_r_l, v_r_l)
            self._set_attrs_after_epoch(epoch_info)

            # Save
            self.save_stats(epoch_info)
            self.save_checkpoint()

            # Log
            self.log_console(epoch_info)
            self.log_tensorboard(epoch_info)

        # Set training flag and save attrs
        self.training_over = True
        self._set_save_attributes()

    def test(self):
        """
        Test SL config.
        """

        self._test_plot_losses()
        self._test_plot_multiple_problems()
        self._test_plot_progression()
        pass

    def save_checkpoint(self):
        """
        Save checkpoint.
        """

        # Save attributes
        self._set_save_attributes()

        # Check if validation loss improved
        if self.model_curr_loss < self.model_best_loss:
            # Log
            self.console.print(f"\n[red bold underline]Validation loss decreased from "
                               f"{self.model_best_loss:.2e} to {self.model_curr_loss:.2e}... Saving model...[/]")

            # Save model
            self._set_save_model()

    def save_stats(self, epoch_info: 'SLEpochInfo') -> None:
        """
        Save train statistics.
        :param epoch_info: current epoch info.
        :return: None.
        """

        # Get locals
        ep = epoch_info.epoch_n
        t_l = epoch_info.train_loss
        v_l = epoch_info.val_loss
        t_kl_l = epoch_info.train_kl_loss
        v_kl_l = epoch_info.val_kl_loss
        t_r_l = epoch_info.train_recon_loss
        v_r_l = epoch_info.val_recon_loss
        b = self.cfg.sl.train.batch_size

        # Compute mean loss = mean(epoch_losses) / batch_size
        mean_train_loss = np.array(t_l).mean() / b
        mean_val_loss = np.array(v_l).mean() / b
        mean_train_kl_loss = np.array(t_kl_l).mean() / b
        mean_val_kl_loss = np.array(v_kl_l).mean() / b
        mean_train_recon_loss = np.array(t_r_l).mean() / b
        mean_val_recon_loss = np.array(v_r_l).mean() / b

        # Save to object
        self.stats.train_loss.append(mean_train_loss)
        self.stats.val_loss.append(mean_val_loss)
        self.stats.train_kl_loss.append(mean_train_kl_loss)
        self.stats.val_kl_loss.append(mean_val_kl_loss)
        self.stats.train_recon_loss.append(mean_train_recon_loss)
        self.stats.val_recon_loss.append(mean_val_recon_loss)

        # Generate states and save
        for idx in range(self.cfg.sl.val.n_maps):
            # Get sample
            sample = self.dataset_val[idx]
            y = sample['y']

            # Generate samples
            states = self.get_samples(y, self.cfg.sl.val.n_states, np.random.default_rng())

            # Append
            self.stats.arr_epoch_idx_state_dim[ep, idx, :] = states

        # Save object in file
        self.stats.save_to_file(self.path_stats)

    def log_console(self, epoch_info: 'SLEpochInfo') -> None:
        """
        Log to console
        :param epoch_info: current epoch info.
        :return: None.
        """

        # Get locals
        epoch_n = epoch_info.epoch_n
        t_l = epoch_info.train_loss
        v_l = epoch_info.val_loss
        b = self.cfg.sl.train.batch_size

        # Compute mean loss = mean(epoch_losses) / batch_size
        mean_train_loss = np.array(t_l).mean() / b
        mean_val_loss = np.array(v_l).mean() / b

        self.console.print(
            f"\n[blue bold underline]Epoch:[/blue bold underline] [blue]{epoch_n}[/blue]\t"
            f"[cyan bold underline]Train loss:[/cyan bold underline] [cyan]{mean_train_loss:.2e}[/cyan]\t"
            f"[green bold underline]Val loss:[/green bold underline] [green]{mean_val_loss:.2e}[/green]"
        )

    def log_tensorboard(self, epoch_info: 'SLEpochInfo') -> None:
        """
        Log info to tensorboard.
        :param epoch_info: current epoch info.
        :return: None
        """

        epoch_n = epoch_info.epoch_n
        t_l = epoch_info.train_loss
        v_l = epoch_info.val_loss
        t_kl_l = epoch_info.train_kl_loss
        v_kl_l = epoch_info.val_kl_loss
        t_r_l = epoch_info.train_recon_loss
        v_r_l = epoch_info.val_recon_loss
        b = self.cfg.sl.train.batch_size

        # Compute mean loss = mean(epoch_losses) / batch_size
        mean_train_loss = np.array(t_l).mean() / b
        mean_val_loss = np.array(v_l).mean() / b
        mean_train_kl_loss = np.array(t_kl_l).mean() / b
        mean_val_kl_loss = np.array(v_kl_l).mean() / b
        mean_train_recon_loss = np.array(t_r_l).mean() / b
        mean_val_recon_loss = np.array(v_r_l).mean() / b

        # Write losses to tensorboard
        loss = {'train': mean_train_loss, 'val': mean_val_loss}
        self.tb.add_scalars("loss", loss, epoch_n)
        kl_loss = {'train': mean_train_kl_loss, 'val': mean_val_kl_loss}
        self.tb.add_scalars("kl_loss", kl_loss, epoch_n)
        recon_loss = {'train': mean_train_recon_loss, 'val': mean_val_recon_loss}
        self.tb.add_scalars("recon_loss", recon_loss, epoch_n)

        # Check generation of states
        self.tb.add_figure("generated_poses", self._get_tb_plot_predictions(), epoch_n)

    def get_conditions(self, map_num: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Get conditions from validation dataset (sample with index 'map_num').
        :return: obstacles, initial state and final state
        """

        # Maps are grouped with different samples
        # idx = 20 * map_num
        idx = map_num

        # Get whole condition vector
        y = self.dataset_val[idx]['y']

        # Separate
        obstacles = y[0:self.cfg.sl.dim.obstacle]  # type: t.Tensor
        state_i = y[self.cfg.sl.dim.obstacle:(self.cfg.sl.dim.obstacle + self.cfg.sl.dim.state)]  # type: t.Tensor
        state_f = y[-self.cfg.sl.dim.state:]  # type: t.Tensor

        # Convert to numpy array
        obstacles = obstacles.cpu().numpy()  # type: np.ndarray
        state_i = state_i.cpu().numpy()  # type: np.ndarray
        state_f = state_f.cpu().numpy()  # type: np.ndarray

        return obstacles, state_i, state_f

    def get_samples(self, y: np.ndarray | t.Tensor | int, n_samples: int | None, rng: Generator) -> np.ndarray:
        """
        Generate states.
        """

        # Case y is int we get sample from validation set
        if isinstance(y, int):
            y = self.dataset_val[y]['y']

        # First replicate condition vector along the first axis and move to cuda
        if isinstance(y, np.ndarray):
            y = t.tensor(y).repeat((n_samples, 1)).cuda()
        else:
            y = y.repeat((n_samples, 1)).cuda()

        # Turn on eval mode and stop gradient tracking
        self.model.eval()
        with t.no_grad():

            # Sample 'n_samples' from latent vector and concatenate with condition vector
            z = t.tensor(rng.normal(0, 1, (n_samples, self.cfg.sl.dim.latent))).cuda()
            z_and_y = t.cat((z, y), dim=1).float()

            # Decode
            x = self.model.mlp_decoder(z_and_y)

            # Move to cpu and get as numpy array
            x = x.cpu().numpy()

        return x

    def _get_tb_plot_predictions(self):
        """
        Get plot of generated predictions on current epoch
        :return: matplotlib figure to add to tensorboard
        """

        # Get constants (idx is the first validation sample)
        ep = self.epoch
        idx = 0

        # Get states from saved stats
        states = self.stats.arr_epoch_idx_state_dim[ep, idx, :]

        # Initialize map and get plot axis, then get figure from axis
        maps = Map(self.cfg)
        maps.set_add_states(states[:50, :], 'sl')
        fig, ax = maps.get_plot_lines()

        return fig

    def _set_train_on_batch(self, sample: dict) -> Vector3:
        """
        Train model on a batch
        :param sample: batch sample
        :return: total loss, KL loss and reconstruction loss
        """

        # Prepare
        self.model.train()
        self.model_optimizer.zero_grad()

        # Forward
        x, mu, sigma = self.model(sample['x'], sample['y'])

        # Loss and optimize
        loss, kl_loss, recon_loss = self.model.loss(x, sample['x'], mu, sigma)
        loss.backward()
        self.model_optimizer.step()

        return loss.item(), kl_loss.item(), recon_loss.item()

    def _set_validate_on_batch(self, sample: dict) -> Vector3:
        """
        Validate model on a batch.
        :param sample: batch sample.
        :return: total loss, KL loss and reconstruction loss.
        """

        # Prepare model
        self.model.eval()

        # Don't track gradients and get loss
        with t.no_grad():
            x, mu, sigma = self.model(sample['x'], sample['y'])
            loss, kl_loss, recon_loss = self.model.loss(x, sample['x'], mu, sigma)

        return loss.item(), kl_loss.item(), recon_loss.item()

    def _set_attrs_after_epoch(self, epoch_info):
        """
        Set attributes after epoch of training.
        :param epoch_info: current epoch info.
        """

        # Update iteration
        self.epoch = epoch_info.epoch_n

        # Update current loss
        self.model_curr_loss = np.array(epoch_info.val_loss).mean() / self.cfg.sl.train.batch_size

        # Check best loss so far
        if len(self.stats.val_loss) > 1:
            self.model_best_loss = min(self.stats.val_loss)

    def _test_plot_losses(self):
        """
        Plot model training and validation losses.
        """

        # Return if figure exists
        path = self.path_figs / 'sl_erros.png'
        if path.exists():
            return

        # Get figure
        fig, axs = plt.subplots(3, 1, sharex='all', squeeze=True, figsize=(12, 9.6))
        fig.subplots_adjust(hspace=0.25)

        # Plot recon loss
        axs[0].plot(range(1, 50), self.stats.train_recon_loss, label="Treino")
        axs[0].plot(range(1, 50), self.stats.val_recon_loss, label="Validação")
        axs[0].set_title('Erro de Reconstrução')
        axs[0].legend()
        axs[0].grid(True)
        axs[0].set_ylim(None, 0.005)

        # Plot KL loss
        axs[1].plot(range(1, 50), self.stats.train_kl_loss, label="Treino")
        axs[1].plot(range(1, 50), self.stats.val_kl_loss, label="Validação")
        axs[1].set_title('Erro de Divergência KL')
        axs[1].legend()
        axs[1].grid(True)

        # Plot total loss
        axs[2].plot(range(1, 50), self.stats.train_loss, label="Treino")
        axs[2].plot(range(1, 50), self.stats.val_loss, label="Validação")
        axs[2].set_title('Erro Total')
        axs[2].legend()
        axs[2].grid(True)
        axs[2].set_ylim(None, 0.01)
        axs[2].set_xlabel('Época')

        # Save
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_plot_progression(self):
        """
        Plot progression of sample generation.
        """

        # Return if figure exists
        path = self.path_figs / 'sl_progresso.png'
        if path.exists():
            return

        # Constants
        map_num = 23

        # Get problem conditions and change config before creating the map
        obstacles, state_i, state_f = self.get_conditions(map_num)
        cfg = self.cfg.copy(deep=True)
        cfg.maps.general.map_name = 'narrow'
        cfg.maps.narrow.narrow1_pos = obstacles[:2]
        cfg.maps.narrow.narrow2_pos = obstacles[2:]
        cfg.env.car.pose = state4_to_pose3(state_i)
        cfg.env.target.pose = state4_to_pose3(state_f)

        # Initialize figure
        fig, axs = plt.subplots(2, 3, sharex='col', sharey='row', figsize=(12, 9.6))  # type: plt.Figure
        fig.subplots_adjust(hspace=-0.2)
        axs = axs.flatten()

        # Which epochs to plot
        epochs = [1, 9, 19, 29, 39, 49]

        # Plot
        for i, epoch in enumerate(epochs):

            # Get states
            states = self.stats.arr_epoch_idx_state_dim[epoch, map_num, :15]

            # Create map and add states
            mapp = Map(cfg)
            mapp.set_add_states(states, 'sl')
            mapp.set_add_states(state_i, 'init')
            mapp.set_add_states(state_f, 'final')
            mapp.get_plot_lines(axs[i])

            # Add title
            axs[i].set_title(f'Época {epoch}')

            # Axes labels
            if i in [0, 3]:
                axs[i].set_ylabel('y [m]')
            if i in [3, 4, 5]:
                axs[i].set_xlabel('x [m]')

        # Save
        fig.savefig(str(path), dpi=600, bbox_inches='tight')

    def _test_plot_multiple_problems(self):
        """
        Get CVAE solution on test problems.
        """

        # Return if figure exists
        path = self.path_figs / 'sl_diversos_problemas.png'
        if path.exists():
            return

        # Initialize figure
        fig, axs = plt.subplots(2, 3, sharex='col', sharey='row', figsize=(12, 9.6))  # type: plt.Figure
        fig.subplots_adjust(hspace=-0.2)
        axs = axs.flatten()

        # Get problems
        problems = [20, 40, 60, 80, 100, 120]

        # Plot
        for i, p in enumerate(problems):

            # Get conditions
            obstacles, state_i, state_f = self.get_conditions(p)
            cfg = self.cfg.copy(deep=True)
            cfg.maps.general.map_name = 'narrow'
            cfg.maps.narrow.narrow1_pos = obstacles[:2]
            cfg.maps.narrow.narrow2_pos = obstacles[2:]
            # cfg.env.car.pose = state4_to_pose3(state_i)
            # cfg.env.target.pose = state4_to_pose3(state_f)
            cfg.env.car.pose = [state_i[0], state_i[1], np.arctan2(state_i[3], state_i[2])]
            cfg.env.target.pose = [state_f[0], state_f[1], np.arctan2(state_f[3], state_f[2])]

            # Get samples
            condition = np.concatenate([obstacles, state_i, state_f])
            states = self.get_samples(condition, 15, np.random.default_rng(0))

            # States are in the form [x y cos sin] but map needs [x y sin cos]
            states = state4_sl_to_state4_rl(states)
            state_i = state4_sl_to_state4_rl(state_i)
            state_f = state4_sl_to_state4_rl(state_f)

            # Create map and add states
            mapp = Map(cfg)
            mapp.set_add_states(states, 'sl')
            mapp.set_add_states(state_i, 'init')
            mapp.set_add_states(state_f, 'final')
            mapp.get_plot_lines(axs[i])

            # Add title
            axs[i].set_title(f'Problema {i+1}')

            # Axes labels
            if i in [0, 3]:
                axs[i].set_ylabel('y [m]')
            if i in [3, 4, 5]:
                axs[i].set_xlabel('x [m]')

        # Save
        fig.savefig(path, dpi=600, bbox_inches='tight')

    def _setup(self):
        """
        Set up.
        """

        # Set up base algorithm settings
        self._setup_paths()
        self._setup_folders()
        self._setup_save_config_to_file()
        self._setup_attrs_to_save()
        self._setup_init_base_attrs()
        self._setup_init_model()
        self._setup_checkpoint()
        self._setup_stats()
        self._setup_console()
        self._setup_tensorboard()

        # Set up SL settings
        self._setup_dataset()
        self._setup_dataloader()
        self._setup_cuda()

    def _setup_checkpoint(self):
        """
        Set up checkpoint.
        """

        # Check if there is a checkpoint
        self.load_checkpoint = self._get_checkpoint_exists_indicator()

        # Load model and attributes if it does
        if self.load_checkpoint:
            self._set_load_attributes()
            self._set_load_model()

    def _setup_cuda(self):
        """
        Set up cuda.
        """

        self.model.to("cuda")

    def _setup_init_model(self):
        """
        Set up CVAE model and optimizer.
        """

        self.model = CVAE(self.cfg)
        self.model_optimizer = Adam(self.model.parameters(), lr=self.cfg.sl.train.lr)

    def _setup_dataloader(self):
        """
        Set up train and validation loaders.
        """

        self.dataloader_train = DataLoader(
            self.dataset_train, self.cfg.sl.train.batch_size, pin_memory=False
        )
        self.dataloader_val = DataLoader(
            self.dataset_val, self.cfg.sl.train.batch_size, pin_memory=False
        )

    def _setup_dataset(self):
        """
        Set up train and validation datasets.
        """

        self.dataset_train = NarrowCVAEDataset(cfg=self.cfg, train=True)
        self.dataset_val = NarrowCVAEDataset(cfg=self.cfg, train=False)

    def _setup_stats(self):
        """
        Set up stats object.
        """

        if self.load_checkpoint:
            self.stats = SLStats.load_from_file(self.path_stats)
        else:
            self.stats = SLStats.new(self.cfg)
            self.stats.save_to_file(self.path_stats)



"""
Curiosity modules for unsupervised exploration
"""

import mrl
import numpy as np
from mrl.replays.online_her_buffer import OnlineHERBuffer
from mrl.utils.misc import softmax, AttrDict
from sklearn.neighbors import KernelDensity
from collections import deque


def generate_overshooting_goals(num_proposals, step_amount, direct_overshoots, base_goal):
  base_proposals = np.array([base_goal, base_goal + step_amount])
  if direct_overshoots:
    return base_proposals
  additional_proposals = base_goal[None] + np.random.uniform(
      -1.5, 1.5, (num_proposals - 2, step_amount.shape[0])) * step_amount[None]
  return np.concatenate((base_proposals, additional_proposals), 0)


class AchievedGoalCuriosity(mrl.Module):
  """
    For goal agents only. This module assumes the replay buffer maintains an achieved goal buffer;
    To decide on goals to pursue during exploration, the module samples goals from the achieved goal
    buffer, and chooses the highest scoring (see below) viable (per q-function) goal.  
  """
  def __init__(self, num_sampled_ags=500, max_steps=50, keep_dg_percent=-1e-1, randomize=False, use_qcutoff=True):
    super().__init__('ag_curiosity',
                     required_agent_modules=['env', 'replay_buffer', 'actor', 'critic'],
                     locals=locals())
    self.num_sampled_ags = num_sampled_ags
    self.max_steps = max_steps  #TODO: have this be learned from past trajectories?
    self.keep_dg_percent = keep_dg_percent
    self.randomize = randomize
    self.use_qcutoff = use_qcutoff

  def _setup(self):
    assert isinstance(self.replay_buffer, OnlineHERBuffer)
    assert self.env.goal_env

    self.n_envs = self.env.num_envs
    self.current_goals = None
    self.replaced_goal = np.zeros((self.env.num_envs, ))

    # setup cutoff
    if self.config.gamma < 1.:
      r = min(self.config.gamma, 0.99)
      self.min_min_cutoff = -(1 - r**(self.max_steps * 0.8)) / (1 - r)
    else:
      self.min_min_cutoff = -self.max_steps * 0.8
    self.min_cutoff = max(self.config.initial_cutoff, self.min_min_cutoff)
    self.cutoff = self.min_cutoff

    # go explore + success accounting
    self.go_explore = np.zeros((self.n_envs, 1), dtype=np.float32)
    self.is_success = np.zeros((self.n_envs, 1), dtype=np.float32)
    self.successes_deque = deque(maxlen=10)  # for dynamic cutoff
    self.successes = []

  def _manage_resets_and_success_behaviors(self, experience, close):
    """ Manage (1) end of trajectory, (2) early resets, (3) go explore and overshot goals """
    reset_idxs, overshooting_idxs, overshooting_proposals = [], [], []

    for i, over in enumerate(experience.trajectory_over):
      if over:  # if over update it
        self.current_goals[i] = experience.reset_state['desired_goal'][i]
        self.replaced_goal[i] = 0.
        if np.random.random() < (self.go_explore[i] * self.config.go_reset_percent):
          reset_idxs.append(i)

      if not over and close[i]:  # if not over and success, modify go_explore; maybe overshoot goal?
        self.is_success[i] += 1.
        self.go_explore[i] += 1.

        if not self.config.get('never_done') and np.random.random() < self.config.overshoot_goal_percent:
          step_amount = experience.next_state['achieved_goal'][i] - experience.state['achieved_goal'][i]
          overshooting_idxs.append(i)
          overshooting_proposals.append(
              generate_overshooting_goals(self.num_sampled_ags, step_amount, self.config.direct_overshoots,
                                          self.current_goals[i]))

    return reset_idxs, overshooting_idxs, np.array(overshooting_proposals)

  def _overshoot_goals(self, experience, overshooting_idxs, overshooting_proposals):
    #score the proposals
    num_proposals = overshooting_proposals.shape[1]
    num_idxs = len(overshooting_idxs)
    states = np.tile(experience.reset_state['observation'][overshooting_idxs, None, :], (1, num_proposals, 1))
    states = np.concatenate((states, overshooting_proposals), -1).reshape(num_proposals * num_idxs, -1)

    bad_q_idxs, q_values = [], None
    if self.use_qcutoff:
      q_values = self.compute_q(states)
      q_values = q_values.reshape(num_idxs, num_proposals)
      bad_q_idxs = q_values < self.cutoff
    goal_values = self.score_goals(overshooting_proposals, AttrDict(q_values=q_values, states=states))

    if self.config.dg_score_multiplier > 1. and self.dg_kde.ready:
      dg_scores = self.dg_kde.evaluate_log_density(overshooting_proposals.reshape(num_proposals * num_idxs, -1))
      dg_scores = dg_scores.reshape(num_idxs, num_proposals)
      goal_values[dg_scores > -np.inf] *= self.config.dg_score_multiplier

    goal_values[bad_q_idxs] = q_values[bad_q_idxs] * -1e-8

    chosen_idx = np.argmin(goal_values, axis=1)
    chosen_idx = np.eye(num_proposals)[chosen_idx]  # shape(sampled_ags) = n_envs x num_proposals
    chosen_ags = np.sum(overshooting_proposals * chosen_idx[:, :, None], axis=1)  # n_envs x goal_feats

    for idx, goal in zip(overshooting_idxs, chosen_ags):
      self.current_goals[idx] = goal
      self.replaced_goal[idx] = 1.

  def _process_experience(self, experience):
    """Curiosity module updates the desired goal depending on experience.trajectory_over"""
    ag_buffer = self.replay_buffer.buffer.BUFF.buffer_ag

    if self.current_goals is None:
      self.current_goals = experience.reset_state['desired_goal']

    computed_reward = self.env.compute_reward(experience.next_state['achieved_goal'], self.current_goals, 
      {'s':experience.state['observation'], 'ns':experience.next_state['observation']})
    close = computed_reward > -0.5

    # First, manage the episode resets & any special behavior that occurs on goal achievement, like go explore / resets / overshooting
    reset_idxs, overshooting_idxs, overshooting_proposals = self._manage_resets_and_success_behaviors(experience, close)

    if reset_idxs:
      self.train.reset_next(reset_idxs)

    if overshooting_idxs and len(ag_buffer):
      self._overshoot_goals(experience, overshooting_idxs, overshooting_proposals)

    # Now consider replacing the current goals with something else:
    if np.any(experience.trajectory_over) and len(ag_buffer):
      # sample some achieved goals
      sample_idxs = np.random.randint(len(ag_buffer), size=self.num_sampled_ags * self.n_envs)
      sampled_ags = ag_buffer.get_batch(sample_idxs)
      sampled_ags = sampled_ags.reshape(self.n_envs, self.num_sampled_ags, -1)

      # compute the q-values of both the sampled achieved goals and the current goals
      states = np.tile(experience.reset_state['observation'][:, None, :], (1, self.num_sampled_ags, 1))
      states = np.concatenate((states, sampled_ags), -1).reshape(self.num_sampled_ags * self.n_envs, -1)
      states_curr = np.concatenate((experience.reset_state['observation'], self.current_goals), -1)
      states_cat = np.concatenate((states, states_curr), 0)

      bad_q_idxs, q_values = [], None
      if self.use_qcutoff:
        q_values = self.compute_q(states_cat)
        q_values, curr_q = np.split(q_values, [self.num_sampled_ags * self.n_envs])
        q_values = q_values.reshape(self.n_envs, self.num_sampled_ags)

        # Set cutoff dynamically by using intrinsic_success_percent
        if len(self.successes_deque) == 10:
          self.min_cutoff = max(self.min_min_cutoff, min(np.min(q_values), self.min_cutoff))
          intrinsic_success_percent = np.mean(self.successes_deque)
          if intrinsic_success_percent >= self.config.cutoff_success_threshold[1]:
            self.cutoff = max(self.min_cutoff, self.cutoff - 1.)
            self.successes_deque.clear()
          elif intrinsic_success_percent <= self.config.cutoff_success_threshold[0]:
            self.cutoff = max(min(self.config.initial_cutoff, self.cutoff + 1.), self.min_min_cutoff)
            self.successes_deque.clear()

        # zero out the "bad" values. This practically eliminates them as candidates if any goals are viable.
        bad_q_idxs = q_values < self.cutoff
        q_values[bad_q_idxs] *= -1
        min_q_values = np.min(q_values, axis=1, keepdims=True)  # num_envs x1
        q_values[bad_q_idxs] *= -1

      # score the goals -- lower is better
      goal_values = self.score_goals(sampled_ags, AttrDict(q_values=q_values, states=states))

      if self.config.dg_score_multiplier > 1. and self.dg_kde.ready:
        dg_scores = self.dg_kde.evaluate_log_density(sampled_ags.reshape(self.n_envs * self.num_sampled_ags, -1))
        dg_scores = dg_scores.reshape(self.n_envs, self.num_sampled_ags)
        goal_values[dg_scores > -np.inf] *= self.config.dg_score_multiplier

      if q_values is not None:
        goal_values[bad_q_idxs] = q_values[bad_q_idxs] * -1e-8

      if self.randomize:  # sample proportional to the absolute score
        abs_goal_values = np.abs(goal_values)
        normalized_values = abs_goal_values / np.sum(abs_goal_values, axis=1, keepdims=True)
        chosen_idx = (normalized_values.cumsum(1) > np.random.rand(normalized_values.shape[0])[:, None]).argmax(1)
      else:  # take minimum
        chosen_idx = np.argmin(goal_values, axis=1)

      chosen_idx = np.eye(self.num_sampled_ags)[chosen_idx]  # shape(sampled_ags) = n_envs x num_sampled_ags
      if q_values is not None:
        chosen_q_val = (chosen_idx * q_values).sum(axis=1, keepdims=True)
      chosen_ags = np.sum(sampled_ags * chosen_idx[:, :, None], axis=1)  # n_envs x goal_feats

      # replace goal always when first_visit_succ (relying on the dg_score_multiplier to dg focus), otherwise
      # we are going to transition into the dgs using the ag_kde_tophat
      if hasattr(self, 'curiosity_alpha'):
        if self.use_qcutoff:
          replace_goal = np.logical_or((np.random.random((self.n_envs, 1)) > self.curiosity_alpha.alpha),
                                       curr_q < self.cutoff).astype(np.float32)
        else:
          replace_goal = (np.random.random((self.n_envs, 1)) > self.curiosity_alpha.alpha).astype(np.float32)

      else:
        replace_goal = np.ones((self.n_envs, 1), dtype=np.float32)

      # sometimes keep the desired goal anyways
      replace_goal *= (np.random.uniform(size=[self.n_envs, 1]) > self.keep_dg_percent).astype(np.float32)

      new_goals = replace_goal * chosen_ags + (1 - replace_goal) * self.current_goals

      if hasattr(self, 'logger') and len(self.successes) > 50:
        if q_values is not None:
          self.logger.add_histogram('Explore/Goal_q', replace_goal * chosen_q_val + (1 - replace_goal) * curr_q)
        self.logger.add_scalar('Explore/Intrinsic_success_percent', np.mean(self.successes))
        self.logger.add_scalar('Explore/Cutoff', self.cutoff)
        self.successes = []

      replace_goal = replace_goal.reshape(-1)

      for i in range(self.n_envs):
        if experience.trajectory_over[i]:
          self.successes.append(float(self.is_success[i, 0] >= 1.))  # compromise due to exploration
          self.successes_deque.append(float(self.is_success[i, 0] >= 1.))  # compromise due to exploration
          self.current_goals[i] = new_goals[i]
          if replace_goal[i]:
            self.replaced_goal[i] = 1.
          self.go_explore[i] = 0.
          self.is_success[i] = 0.

  def compute_q(self, numpy_states):
    states = self.torch(numpy_states)
    max_actions = self.actor(states)
    if isinstance(max_actions, tuple):
      max_actions = max_actions[0]
    return self.numpy(self.critic(states, max_actions))

  def relabel_state(self, state):
    """Should be called by the policy module to relabel states with intrinsic goals"""
    if self.current_goals is None:
      return state

    return {
        'observation': state['observation'],
        'achieved_goal': state['achieved_goal'],
        'desired_goal': self.current_goals
    }

  def score_goals(self, sampled_ags, info):
    """ Lower is better """
    raise NotImplementedError  # SUBCLASS THIS!

  def save(self, save_folder):
    self._save_props(['cutoff', 'min_cutoff'], save_folder)  #can restart keeping track of successes / go explore

  def load(self, save_folder):
    self._load_props(['cutoff', 'min_cutoff'], save_folder)


class QAchievedGoalCuriosity(AchievedGoalCuriosity):
  """
  Scores goals by the Q values (lower is better)
  """
  def score_goals(self, sampled_ags, info):
    scores = np.copy(info.q_values)
    max_score = np.max(scores)
    if max_score > 0:
      scores -= max_score  # so all scores negative

    return scores


class SuccessAchievedGoalCuriosity(AchievedGoalCuriosity):
  """
  Scores goals based on success prediction by a goal discriminator module.
  """
  def _setup(self):
    super()._setup()
    self.use_qcutoff = False

  def score_goals(self, sampled_ags, info):

    # sampled_ags is np.array of shape NUM_ENVS x NUM_SAMPLED_GOALS (both arbitrary)
    num_envs, num_sampled_ags = sampled_ags.shape[:2]

    scores = self.success_predictor(info.states).reshape(num_envs, num_sampled_ags)  # these are predicted success %
    scores = -0.5 + np.abs(scores - 0.5)  # rank by distance to 0.5, lower is closer to 0.5

    return scores


class DensityAchievedGoalCuriosity(AchievedGoalCuriosity):
  """
  Scores goals by their densities (lower is better), using KDE to estimate

  Note on bandwidth: it seems bandwith = 0.1 works pretty well with normalized samples (which is
  why we normalize the ags).
  """
  def __init__(self, density_module='ag_kde', interest_module='ag_interest', alpha=-1.0, **kwargs):
    super().__init__(**kwargs)
    self.alpha = alpha
    self.density_module = density_module
    self.interest_module = interest_module

  def _setup(self):
    assert hasattr(self, self.density_module)
    super()._setup()

  def score_goals(self, sampled_ags, info):
    """ Lower is better """
    density_module = getattr(self, self.density_module)
    if not density_module.ready:
      density_module._optimize(force=True)
    interest_module = None
    if hasattr(self, self.interest_module):
      interest_module = getattr(self, self.interest_module)
      if not interest_module.ready:
        interest_module = None

    # sampled_ags is np.array of shape NUM_ENVS x NUM_SAMPLED_GOALS (both arbitrary)
    num_envs, num_sampled_ags = sampled_ags.shape[:2]

    # score the sampled_ags to get log densities, and exponentiate to get densities
    flattened_sampled_ags = sampled_ags.reshape(num_envs * num_sampled_ags, -1)
    sampled_ag_scores = density_module.evaluate_log_density(flattened_sampled_ags)
    if interest_module:
      # Interest is ~(det(feature_transform)), so we subtract it  in order to add ~(det(inverse feature_transform)) for COV.
      sampled_ag_scores -= interest_module.evaluate_log_interest(flattened_sampled_ags)  # add in log interest
    sampled_ag_scores = sampled_ag_scores.reshape(num_envs, num_sampled_ags)  # these are log densities

    # Take softmax of the alpha * log density.
    # If alpha = -1, this gives us normalized inverse densities (higher is rarer)
    # If alpha < -1, this skews the density to give us low density samples
    normalized_inverse_densities = softmax(sampled_ag_scores * self.alpha)
    normalized_inverse_densities *= -1.  # make negative / reverse order so that lower is better.

    return normalized_inverse_densities


class EntropyGainScoringGoalCuriosity(AchievedGoalCuriosity):
  """
  Scores goals by their expected entropy gain (higher is better), using KDE to estimate
  current density and another KDE to estimate the joint likelihood of achieved goal 
  given behavioural goal.
  """
  def __init__(self, **kwargs):
    super().__init__(**kwargs)

  def _setup(self):
    assert hasattr(self, 'bg_kde')
    assert hasattr(self, 'ag_kde')
    assert hasattr(self, 'bgag_kde')
    assert hasattr(self, 'replay_buffer')
    super()._setup()

  def score_goals(self, sampled_ags, info):
    """ Higher entropy gain is better """
    if not self.ag_kde.ready:
      self.ag_kde._optimize(force=True)

    if not self.bg_kde.ready:
      self.bg_kde._optimize(force=True)

    if not self.bgag_kde.ready:
      self.bgag_kde._optimize(force=True)

    # sampled_ags is np.array of shape NUM_ENVS x NUM_SAMPLED_GOALS (both arbitrary)
    num_envs, num_sampled_ags = sampled_ags.shape[:2]

    # Get sample of predicted achieved goal from mixture density network
    candidate_bgs = sampled_ags.reshape(num_envs * num_sampled_ags, -1)

    # Reuse the candidate bgs as potential ags
    # Note: We are using a sliding window to reuse sampled_ags as the potential ag for each bg
    # Prior that each bgs has one ag that is identical to bg, i.e. that it reaches the bg.
    num_ags = 10  # TODO: Not make it hard coded
    indexer = np.arange(num_envs * num_sampled_ags).reshape(-1, 1) + np.arange(num_ags).reshape(1, -1)
    indexer %= num_envs * num_sampled_ags  # To wrap around to the beginning
    ags_samples = np.concatenate(
        [candidate_bgs[indexer[i]][np.newaxis, :, :] for i in range(num_envs * num_sampled_ags)], axis=0)

    candidate_bgs_repeat = np.repeat(candidate_bgs[:, np.newaxis, :], num_ags,
                                     axis=1)  # Shape num_envs*num_sampled_ags, num_ags, dim
    joint_candidate_bgags = np.concatenate([candidate_bgs_repeat, ags_samples], axis=-1)
    joint_candidate_bgags = joint_candidate_bgags.reshape(num_envs * num_sampled_ags * num_ags, -1)

    # score the sampled_ags to get log densities, and exponentiate to get densities
    joint_candidate_score = self.bgag_kde.evaluate_log_density(joint_candidate_bgags)
    joint_candidate_score = joint_candidate_score.reshape(num_envs * num_sampled_ags,
                                                          num_ags)  # these are log densities

    candidate_bgs_score = self.bg_kde.evaluate_log_density(
        candidate_bgs_repeat.reshape(num_envs * num_sampled_ags * num_ags, -1))
    candidate_bgs_score = candidate_bgs_score.reshape(num_envs * num_sampled_ags, num_ags)  # these are log densities
    cond_candidate_score = joint_candidate_score - candidate_bgs_score
    cond_candidate_score = softmax(cond_candidate_score, axis=1)

    # Compute entropy gain for the predicted achieved goal
    beta = 1 / len(self.replay_buffer.buffer)
    sampled_ag_entr_new = self.ag_kde.evaluate_elementwise_entropy(candidate_bgs, beta=beta)
    sampled_ag_entr_old = self.ag_kde.evaluate_elementwise_entropy(candidate_bgs, beta=0.)
    sampled_ag_entr_gain = sampled_ag_entr_new - sampled_ag_entr_old
    sampled_ag_entr_gain /= beta  # Normalize by beta # TODO: Get rid of this part if not necessary
    sampled_ag_entr_gain = np.concatenate(
        [sampled_ag_entr_gain[indexer[i]][np.newaxis, :] for i in range(num_envs * num_sampled_ags)], axis=0)
    sampled_ag_entr_gain *= cond_candidate_score
    sampled_ag_entr_gain = sampled_ag_entr_gain.mean(axis=1)

    scores = sampled_ag_entr_gain.reshape(num_envs, num_sampled_ags)
    scores *= -1.  # make negative / reverse order so that lower is better.

    return scores


class CuriosityAlphaMixtureModule(mrl.Module):
  """
    For curiosity agents; this module approximates alpha = (1 / (1 + KL)) using the ag_kde density estimator for p_ag.
  """
  def __init__(self, optimize_every=100):
    super().__init__('curiosity_alpha',
                     required_agent_modules=['ag_curiosity', 'ag_kde', 'replay_buffer'],
                     locals=locals())
    self.samples = None
    self.bandwidth = None
    self.kernel = None
    self.kde = None
    self.fitted_kde = None
    self._alpha = 0.
    self._beta = -3.
    self.step = 0
    self.optimize_every = optimize_every

  def _setup(self):
    self.samples = self.ag_kde.samples
    self.bandwidth = self.ag_kde.bandwidth
    self.kernel = self.ag_kde.kernel
    self.kde = KernelDensity(kernel=self.kernel, bandwidth=self.bandwidth)
    if 'curiosity_beta' in self.config:
      self._beta = self.config.curiosity_beta

  @property
  def alpha(self):
    return self._alpha

  def _optimize(self):
    buffer = self.replay_buffer.buffer.BUFF['buffer_dg']
    self.step += 1

    if self.step % self.optimize_every == 0 and len(buffer):

      # Fit the DG KDE
      num_samples = 1000
      sample_idxs = np.random.randint(len(buffer), size=num_samples)
      kde_samples = buffer.get_batch(sample_idxs)
      kde_samples = (kde_samples - self.ag_kde.kde_sample_mean) / self.ag_kde.kde_sample_std
      self.fitted_kde = self.kde.fit(kde_samples)

      # Now compute alpha
      s = kde_samples
      log_p_dg = self.fitted_kde.score_samples(s)
      log_p_ag = self.ag_kde.fitted_kde.score_samples(s)
      self._alpha = 1. / max((self._beta + np.mean(log_p_dg) - np.mean(log_p_ag)), 1.)

      # Occasionally log the alpha
      self.logger.add_scalar('Explore/curiosity_alpha', self._alpha, log_every=500)
      self.logger.add_tabular('Curiosity_alpha', self._alpha)

  def save(self, save_folder):
    self._save_props(['kde', 'samples', 'bandwidth', 'kernel'], save_folder)

  def load(self, save_folder):
    self._load_props(['kde', 'samples', 'bandwidth', 'kernel'], save_folder)



"""
These are schedules for exploration. Passed to the random_process.
"""


#######################################################################
# Copyright (C) 2017 Shangtong Zhang(zhangshangtong.cpp@gmail.com)    #
# Permission given to modify the code as long as you keep this        #
# declaration at the top                                              #
#######################################################################

class ConstantSchedule:
    def __init__(self, val):
        self.val = val

    def __call__(self, steps=1):
        return self.val

class LinearSchedule:
    def __init__(self, start, end=None, steps=None):
        if end is None:
            end = start
            steps = 1
        self.inc = (end - start) / float(steps)
        self.current = start
        self.end = end
        if end > start:
            self.bound = min
        else:
            self.bound = max

    def __call__(self, steps=1):
        val = self.current
        self.current = self.bound(self.current + self.inc * steps, self.end)
        return val



"""
This is old replay buffer. Does not work with HER. 
Sole advantage is that it does not wait  for the trajectory to finish to add things to the buffer,
  which is important for long Mujoco tasks. 
"""

import mrl
import gym
from mrl.replays.core.replay_buffer import ReplayBuffer as Buffer
from typing import Optional
import numpy as np
import torch
import pickle
import os

class OldReplayBuffer(mrl.Module):

  def __init__(self):
    """
    A standard replay buffer (no prioritization / fancy stuff). 
    """

    super().__init__('replay_buffer', required_agent_modules=['env'], locals=locals())

    self.size = None
    self.goal_space = None
    self.hindsight_buffer = None
    self.buffer = None
    self.save_buffer = None

  def _setup(self):
    self.size = self.config.replay_size

    env = self.env
    if type(env.observation_space) == gym.spaces.Dict:
      observation_space = env.observation_space.spaces["observation"]
      self.goal_space = env.observation_space.spaces["desired_goal"]
      raise NotImplementedError("This buffer no longer supports goal spaces; use OnlineHERBuffer")
    else:
      observation_space = env.observation_space

    items = [("state", observation_space.shape),
             ("action", env.action_space.shape), ("reward", (1,)),
             ("next_state", observation_space.shape), ("done", (1,))]

    self.buffer = Buffer(self.size, items)

  def _process_experience(self, experience):
    if getattr(self, 'logger'):
      self.logger.add_tabular('Replay buffer size', len(self.buffer))
    done = np.expand_dims(experience.done, 1) # format for replay buffer
    reward = np.expand_dims(experience.reward, 1) # format for replay buffer
    action = experience.action

    state = experience.state
    next_state = experience.next_state
    self.buffer.add_batch(state, action, reward, next_state, done)
  
  def sample(self, batch_size, to_torch=True):
    states, actions, rewards, next_states, dones = self.buffer.sample(batch_size)
    gammas = self.config.gamma * (1-dones)
    
    if hasattr(self, 'state_normalizer'):
      states = self.state_normalizer(states, update=False).astype(np.float32)
      next_states = self.state_normalizer(next_states, update=False).astype(np.float32)

    if to_torch:
      return (self.torch(states), self.torch(actions),
            self.torch(rewards), self.torch(next_states),
            self.torch(gammas))
    else:
      return (states, actions, rewards, next_states, gammas)
        
  def __len__(self):
    return len(self.buffer)

  def save(self, save_folder):
    if self.config.save_replay_buf or self.save_buffer:
      state = self.buffer._get_state()
      with open(os.path.join(save_folder, "{}.pickle".format(self.module_name)), 'wb') as f:
        pickle.dump(state, f)

  def load(self, save_folder):
    load_path = os.path.join(save_folder, "{}.pickle".format(self.module_name))
    if os.path.exists(load_path):
      with open(load_path, 'rb') as f:
        state = pickle.load(f)
      self.buffer._set_state(state)
    else:
      self.logger.log_color('###############################################################', '', color='red')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='cyan')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='red')
      self.logger.log_color('WARNING', 'Replay buffer is not being loaded / was not saved.', color='yellow')
      self.logger.log_color('###############################################################', '', color='red')



from typing import Any

from pydantic import BaseModel
import yaml

from rrt_ml.utilities.hints import *
from rrt_ml.utilities.paths import *


class MasterConfig(BaseModel):

    class _env(BaseModel):

        class _car(BaseModel):
            axis_dist: float = 0.325
            f_val: int = 10
            phi_dot_max: float = 20.0
            phi_max: float = 0.7
            pose: float | list[float, float, float] = 1.0
            reset_z: float = 0.001
            wheel_dist: float = 0.2
            wheel_radius: float = 0.05
            v_dot_max: float = 1.0
            v_max: float = 1.0

        class _general(BaseModel):
            gui: bool = False
            max_timestep: int = 100
            seed: int = 0
            stats: bool = False

        class _reward(BaseModel):
            epsilon: float = -0.10
            p_val: float = 0.5
            weights: list[float, float] = [0.75, 0.15]

        class _target(BaseModel):
            pose: float | list[float, float, float] = 1.0
            reset_z: float = 0.0

        class _test(BaseModel):

            class _general(BaseModel):
                ghost_interval: int = 90

            class _narrow(BaseModel):
                cam_dist: float = 8.0
                cam_target: list[float, float, float] = [5, 5, 0]
                img_crop: list[int, int, int, int] = [232, 126, 808, 704]
                img_extent: list[int, int, int, int] = [-1, 11, -1, 11]

            class _none(BaseModel):
                cam_dist: float = 3.35
                cam_target: list[float, float, float] = [0, 0, 0]
                img_crop: list[int, int, int, int] = [176, 71, 864, 759]
                img_extent: list[int, int, int, int] = [-3, 3, -3, 3]

            general: _general = _general()
            narrow: _narrow = _narrow()
            none: _none = _none()

        car: _car = _car()
        general: _general = _general()
        reward: _reward = _reward()
        target: _target = _target()
        test: _test = _test()

    class _general(BaseModel):

        algorithm: str | None = 'rl'
        config_name_or_prefix: str | None = 'default'
        description: str = ""
        is_hyper: bool | None = False
        is_train: bool | None = True

    class _hyperparams(BaseModel):

        class _general(BaseModel):
            name: str = 'h'

        class _rl(BaseModel):

            class _actor(BaseModel):
                l2: list[float, ...] | None = None
                lr: list[float, ...] | None = None
                weight_decay: list[float, ...] | None = None

            class _critic(BaseModel):
                lr: list[float, ...] | None = None

            class _exploration(BaseModel):
                epsilon: list[float, ...] | None = None
                go_epsilon: list[float, ...] | None = None
                go_reset_percent: list[float, ...] | None = None
                initial: list[int, ...] | None = None
                noise: list[float, ...] | None = None
                warm_up: list[int, ...] | None = None

            class _general(BaseModel):
                gamma: list[float, ...] | None = None
                her: list[str, ...] | None = None

            class _net(BaseModel):
                activ: list[str, ...] | None = None
                grad_value_clipping: list[float, ...] | None = None
                layer_norm: list[bool, ...] | None = None
                layers: list[list[int, ...], ...] | None = [[512, 512, 512], [1024, 1024]]

            class _target(BaseModel):
                grad_clip: list[float, ...] | None = None
                update_frac: list[float, ...] | None = None
                update_freq: list[int, ...] | None = None

            class _train(BaseModel):
                batch_size: list[int, ...] | None = [1024, 2048]
                n_timesteps: list[int, ...] | None = None
                optimize_every: list[int, ...] | None = None
                replay_size: list[int, ...] | None = None

            actor: _actor = _actor()
            critic: _critic = _critic()
            exploration: _exploration = _exploration()
            general: _general = _general()
            net: _net = _net()
            target: _target = _target()
            train: _train = _train()

        class _rrt(BaseModel):
            
            class _general(BaseModel):
                add_last_pose: list[bool, ...] | None = None
                condition: list[str | int, ...] | None = None
                controller: list[str, str] | None = ['rl', 'mpc']
                debug: list[bool, ...] | None = None
                delta_t: list[int, ...] | None = None
                distance_fn: list[str, ...] | None = ['rl-ordinal-00', 'rs']
                distance_fn_mode: list[str, ...] | None = None
                epsilon_multiplier: list[float, ...] | None = None
                goal_epsilon: list[float, ...] | None = None
                n_iter: list[int, ...] | None = None
                policy_limitation: list[bool, ...] | None = None
                rewire: list[bool, ...] | None = None
                seed: list[int | None, ...] | None = [i for i in range(9)]

            class _mpc(BaseModel):
                control_cost: list[float, ...] | None = None
                n_horizon: list[int, ...] | None = None
                n_robust: list[int, ...] | None = None
                simulate: list[bool, ...] | None = None
                step_size: list[float, ...] | None = None
                weight_angle: list[float, ...] | None = None
                weight_pos: list[float, ...] | None = None

            class _names(BaseModel):
                rl: list[str, ...] | None = None
                sl: list[str, ...] | None = None

            class _near(BaseModel):
                cutoff_length: list[float, ...] | None = None
                cutoff_timesteps: list[int, ...] | None = None
                max_n_nodes: list[int, ...] | None = None

            class _nearest(BaseModel):
                min_n_timesteps: list[int, ...] | None = None
                min_length: list[float, ...] | None = None

            class _ordinal(BaseModel):  # opt refer to ranges for optuna: [start, stop] or [start, stop, step]
                batch_size: list[int, ...] | None = None
                data_ratio: list[float, ...] | None = None
                layers: list[list[int, ...], ...] | None = None
                max_timesteps: list[int, ...] | None = None
                n_estimators: list[int, ...] | None = None
                opt: list[bool, ...] | None = None
                opt_batch_size: list[list[int, int, int], ...] | None = None
                opt_n_estimators: list[list[int, int, int], ...] | None = None
                opt_n_layers: list[list[int, int], ...] | None = None
                opt_n_units: list[list[int, int, int], ...] | None = None
                use_model_from_experiment: list[str, ...] | None = None

            class _rs(BaseModel):
                curvature: list[float, ...] | None = None
                step_size: list[float, ...] | None = None
                tolerance: list[float, ...] | None = None

            class _sample(BaseModel):
                goal_prob: list[float, ...] | None = None
                learner_prob: list[Any, ...] | None = None

            general: _general = _general()
            mpc: _mpc = _mpc()
            names: _names = _names()
            near: _near = _near()
            nearest: _nearest = _nearest()
            ordinal: _ordinal = _ordinal()
            rs: _rs = _rs()
            sample: _sample = _sample()

        class _sl(BaseModel):

            class _dim(BaseModel):
                latent: list[int, ...] | None = [2, 3]

            class _loss(BaseModel):
                beta: list[float, ...] | None = [0.001, 0.005, 0.01, 0.05, 0.1, 0.5, 1, 5]

            class _net(BaseModel):
                mlp_units: list[list[int, ...], ...] | None = [[256, 256], [512, 512]]

            class _train(BaseModel):
                batch_size: list[int, ...] | None = None
                lr: list[float, ...] | None = [0.001, 0.01]
                n_epochs: list[int, ...] | None = None

            dim: _dim = _dim()
            loss: _loss = _loss()
            net: _net = _net()
            train: _train = _train()

        general: _general = _general()
        rl: _rl = _rl()
        rrt: _rrt = _rrt()
        sl: _sl = _sl()

    class _maps(BaseModel):

        class _arena(BaseModel):
            length: float = 10
            offset: list[float, float] = [0, 0]
            width: float = 10

        class _general(BaseModel):
            map_name: None | str = None

        class _narrow(BaseModel):
            narrow1_pos: list[float, float] = [3.5, 4.5]
            narrow2_pos: list[float, float] = [6.5, 4.5]
            passage_width: float = 1.4
            size: float = 10
            wall_width: float = 0.4

        class _obstacles(BaseModel):
            color: list[float, float, float, float] = [0, 0, 0, 1]
            height: float = 0.05
            mass: float = 1000.0
            reset_z: float = 0.05

        arena: _arena = _arena()
        general: _general = _general()
        narrow: _narrow = _narrow()
        obstacles: _obstacles = _obstacles()

    class _rl(BaseModel):

        class _actor(BaseModel):
            l2: float = 0.1
            lr: float = 0.001
            weight_decay: float = 0.0

        class _critic(BaseModel):
            lr: float = 0.001

        class _exploration(BaseModel):
            epsilon: float = 0.1
            go_epsilon: float = 0.1
            go_reset_percent: float = 0.0
            initial: int = 5000
            noise: float = 0.1
            warm_up: int = 2500

        class _general(BaseModel):
            gamma: float = 0.99
            her: str = 'rfaab_1_4_3_1_1'
            just_use_policy: bool = False

        class _net(BaseModel):
            activ: str = 'gelu'
            grad_value_clipping: float = 5.0
            layer_norm: bool = True
            layers: list[int, int, int] = [512, 512, 512]

        class _target(BaseModel):
            grad_clip: float = 5.0
            update_frac: float = 0.05
            update_freq: int = 40

        class _train(BaseModel):
            batch_size: int = 2048
            n_envs: int = 1
            n_timesteps: int = 1000000
            optimize_every: int = 1
            replay_size: int = 1000000

        class _val(BaseModel):
            gui: bool = False
            stats: bool = True
            interval: int = 10000
            n_envs = 1
            n_episodes = 50
            target_pose: float = 2.0

        actor: _actor = _actor()
        critic: _critic = _critic()
        exploration: _exploration = _exploration()
        general: _general = _general()
        net: _net = _net()
        target: _target = _target()
        train: _train = _train()
        val: _val = _val()

    class _rrt(BaseModel):

        class _general(BaseModel):
            add_last_pose: bool = True  # when simulate reaching, use the last node instead of the random node
            condition: int | str = 'random'  # 'random' generate a new planning problem and integer get from val set
            controller: str = 'rl'  # mpc
            debug: bool = False
            delta_t: int = 160
            distance_fn: str = 'rl-ordinal-00'  # rl-ordinal-xx / rl-sim / rs
            distance_fn_mode: str = 'percent'  # or 'difference'
            epsilon_multiplier: float = 1.0
            goal_epsilon: float = -0.5
            n_iter: int = 500
            policy_limitation: bool = False
            rewire: bool = True
            seed: int | None = 0

        class _mpc(BaseModel):
            control_cost: float = 0.01
            n_horizon: int = 100
            n_robust: int = 4
            simulate: bool = False
            step_size: float = 0.01
            weight_angle: float = 0.05
            weight_pos: float = 1.0

        class _names(BaseModel):
            rl: str = 'best'
            sl: str = 'best'

        class _near(BaseModel):
            cutoff_length: float = 1.0
            cutoff_timesteps: int = 240
            max_n_nodes: int = 5000

        class _nearest(BaseModel):
            min_n_timesteps: int = 0
            min_length: float = 0.0

        class _ordinal(BaseModel):  # opt refer to ranges for optuna: [start, stop] or [start, stop, step]
            batch_size: int = 128
            data_ratio: float = 1
            layers: list[int, ...] = [256, 256, 256]
            max_timesteps: int | None = 48
            n_estimators: int = 10
            opt: bool = False
            opt_batch_size: list[int, int, int] = [16, 512, 16]
            opt_n_estimators: list[int, int, int] = [10, 40, 5]
            opt_n_layers: list[int, int] = [1, 3]
            opt_n_units: list[int, int, int] = [16, 512, 16]
            use_model_from_experiment: None | str = 'default'

        class _rs(BaseModel):
            curvature: float = 2.3
            step_size: float = 0.01
            tolerance: float = 0.01

        class _sample(BaseModel):
            goal_prob: float = 0.05
            learner_prob: float | list[float, float, float] = [1.38, -186.44, 161.44]

        general: _general = _general()
        mpc: _mpc = _mpc()
        names: _names = _names()
        near: _near = _near()
        nearest: _nearest = _nearest()
        ordinal: _ordinal = _ordinal()
        rs: _rs = _rs()
        sample: _sample = _sample()

    class _sl(BaseModel):

        class _dim(BaseModel):
            state: int = 4
            obstacle: int = 4
            latent: int = 4

        class _general(BaseModel):
            data_ratio: float = 0.5

        class _loss(BaseModel):
            beta: float = 0.01

        class _net(BaseModel):
            mlp_units: list[int, ...] = [512, 512]

        class _train(BaseModel):
            batch_size: int = 64
            lr: float = 0.001
            n_epochs: int = 50

        class _val(BaseModel):
            n_maps: int = 30
            n_states: int = 1000

        dim: _dim = _dim()
        general: _general = _general()
        loss: _loss = _loss()
        net: _net = _net()
        train: _train = _train()
        val: _val = _val()

    env: _env = _env()
    general: _general = _general()
    hyperparams: _hyperparams = _hyperparams()
    maps: _maps = _maps()
    rl: _rl = _rl()
    rrt: _rrt = _rrt()
    sl: _sl = _sl()

    def deep_copy_change(self, attr: str, val: Any):
        """
        Deep copy config and change attribute.
        :param attr: string representing nested attribute to change, ex: 'env.general.gui'
        :param val: value to set attribute
        :return: new config with changed value
        """

        # Deep copy config to return and make a shallow copy to reach last config level
        cfg = self.copy(deep=True)
        shallow = cfg

        # Go to last config level
        nested_attrs = attr.split('.')
        for att in nested_attrs[:-1]:
            shallow = shallow.__getattribute__(att)

        # Now set the attribute using shallow copy
        shallow.__setattr__(nested_attrs[-1], val)

        return cfg

    @classmethod
    def load(cls, args):
        """
        Load config from file.
        :return: new config file with overridden attributes
        """

        # Check if we need to load default
        is_default = args.config is None

        # Change name for default config
        if is_default:
            config_name_or_prefix = 'h' if args.hyper else 'default'
        else:
            config_name_or_prefix = args.config

        # Get parameters from cmd line args and path to config
        is_train = args.train
        is_hyper = args.hyper
        match True:
            case args.rl:
                algorithm = 'rl'
            case args.sl:
                algorithm = 'sl'
            case args.rrt:
                algorithm = 'rrt'
            case _:
                raise NotImplementedError

        # Load dict from config to override default config dict
        if not is_default:

            # If config file exists, load it
            path = Paths().cfg(algorithm, config_name_or_prefix, is_hyper)
            if path.exists():
                with open(str(path)) as f:
                    overrider_dict = yaml.safe_load(f)

            # Maybe it doesn't exist but hyperparameter config do
            else:

                # Check if experiment folder exists
                path = Paths().exp(algorithm, config_name_or_prefix)
                if path.exists():

                    # Experiment folder exists, load config from yaml
                    with open(str(path / 'config.yaml'), 'r') as f:
                        overrider_dict = yaml.safe_load(f)

        else:
            overrider_dict = {}
        merged_dict = cls().dict() | overrider_dict

        # Change general attributes of merged config
        merged_cfg = cls(**merged_dict)
        merged_cfg.general.algorithm = algorithm
        merged_cfg.general.config_name_or_prefix = config_name_or_prefix
        merged_cfg.general.is_train = is_train
        merged_cfg.general.is_hyper = is_hyper

        # Post process merged config
        merged_cfg = cls()._set_defaults(merged_cfg)

        # Override again
        merged_dict = merged_cfg.dict() | overrider_dict

        # Override default config with custom config and return it
        return cls()._set_defaults(merged_cfg)

    @classmethod
    def load_from_experiment(cls, alg, name):
        """
        Load config directly from experiment folder.
        :param alg: algorithm name
        :param name: name of the experiment
        """

        # Get path to experiment config
        path = Paths().exp(alg, name) / 'config.yaml'

        # Check existence
        assert path.exists(), "Trying to load config from experiment that doest not exists."
        
        # Get dict
        with open(str(path)) as f:
            overrider_dict = yaml.safe_load(f)
        
        # Get merged dict
        merged_dict = cls().dict() | overrider_dict

        # Change general attributes of merged config
        return cls(**merged_dict)

    @classmethod
    def _set_defaults(cls, config: 'MasterConfig'):
        """
        Change defaults according to algorithm
        :param config: master config.
        :return: config with default parameters changed.
        """

        match config.general.algorithm:

            # Set defaults for SL
            case 'sl':

                # Map should not be 'None' as in RL
                config.maps.general.map_name = 'narrow'

            # Set defaults for RRT
            case 'rrt':

                # Map should not be 'None' as in RL
                config.maps.general.map_name = 'narrow'

                # Pose is fixed (no random resetting)
                config.env.car.pose = [1, 5, 0]
                config.env.target.pose = [9, 5, 0]

                # Must track stats
                config.env.general.stats = True

                # Action repeat is implemented in RRT don't need to change
                config.env.car.f_val = 240
                config.env.general.max_timestep = 10000

                # Don't need envs from RL agent
                config.rl.general.just_use_policy = True

                # For testing we need GUI to take screenshot
                if not config.general.is_train:
                    config.env.general.gui = True

            case _:
                pass

        return config



import numpy as np
import pybullet as p
from matplotlib import pyplot as plt
from scipy import interpolate

from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.configs import *


# Set stats to true to track bullet stats
cfg = MasterConfig()
cfg.env.general.stats = True
cfg.env.general.gui = True
cfg.env.general.max_timestep = 480

# Reset master to track stats
env = CarNavigationBulletEnv(cfg)
env.reset_master()

# Generate info
env.reset()
done = False
while not done:
    obs, reward, done, info = env.step([1, 1])

# Get variables
t = env.stats.bullet.time[0, :50, 0]
v = env.stats.bullet.car_velocities[0, :51, 0]
a = np.diff(v)
v = v[:50]

# Interpolate
fa = interpolate.interp1d(t, a, kind='zero')
fv = interpolate.interp1d(t, v, kind='zero')
x = np.linspace(t[0], t[-1], 10000)
a_x = fa(x)
v_x = fv(x)

# Config plot
plt.rcParams['axes.grid'] = True

# Plot
fig, axs = plt.subplots(1, 2, squeeze=True)
axs[0].plot(x, v_x, label='Velocidade (m/s)')
axs[0].set_xlabel('Tempo (s)')
axs[0].legend()
axs[1].plot(x, a_x, label='Aceleração (m/s^2)', color='red')
axs[1].set_xlabel('Tempo (s)')
axs[1].legend()
fig.savefig('max_velocity.png', dpi=300)
pass




import copy
import time
from pathlib import Path
from typing import Callable, Optional

import matplotlib as mpl
import optuna
import pandas as pd
import seaborn as sns
from coral_pytorch.dataset import corn_label_from_logits
from matplotlib import pyplot as plt
from numpy.random import Generator
from scipy.spatial.distance import euclidean
from sklearn.preprocessing import StandardScaler
from tbparse import SummaryReader
from torch.utils.data import DataLoader
from torchensemble import SnapshotEnsembleRegressor
from torchensemble.utils import io
from torchensemble.utils.logging import set_logger

from rrt_ml.algorithms.base import *
from rrt_ml.algorithms.rl import *
from rrt_ml.algorithms.sl import *
from rrt_ml.environments.car_navigation_bullet_env import *
from rrt_ml.utilities.analytic import *
from rrt_ml.utilities.datasets import *
from rrt_ml.utilities.formulas import *
from rrt_ml.utilities.infos import *
from rrt_ml.utilities.maps import *
from rrt_ml.utilities.misc import *
from rrt_ml.utilities.models import *
from rrt_ml.utilities.paths import *
from rrt_ml.utilities.stats import *


class RRT(Algorithm):
    """
    RRT* algorithm with variations.
    """

    def __init__(self, cfg: MasterConfig):
        """
        Initialize.
        """

        super(RRT, self).__init__(cfg)
        self.controller_fn = None  # type: Optional[Callable[[dict], Vector2]]
        self.cutoff = None  # type: None | float | int
        self.dbg_map = None  # type: None | Map
        self.dbg_fig = None  # type: None | plt.Figure
        self.dbg_ax = None  # type: None | plt.Axes
        self.distance_fn = None  # type: Optional[Callable[[Node | list[Node, ...], Node], float]]
        self.env = None  # type: None | CarNavigationBulletEnv
        self.found_solution = None  # type: None | bool
        self.list_nodes = None  # type: None | list[Node, ...]
        self.list_sl_nodes = None  # type: None | list[Node, ...]
        self.min_distance = None  # type: None | float | int
        self.mpcrs = None  # type: None | MPCRS
        self.node_final = None  # type: None | Node
        self.node_init = None  # type: None | Node
        self.ordinal = None  # type: None | SnapshotEnsembleRegressor
        self.ordinal_range = None  # type: None | list[int, int]
        self.ordinal_scaler = None  # type: None | StandardScaler
        self.rng = None  # type: None | Generator
        self.rl = None  # type: None | RL
        self.rl_action_repeat = None  # type: None | int
        self.solution_simulation_info = None  # type: None | RRTSolutionInfo

        self._setup()

    def train(self):
        """
        Build RRT tree.
        """

        while len(self.list_nodes) < self.cfg.rrt.general.n_iter:

            # Sample final goal or leaner goal
            node_rand = self._get_random_node()

            # Check if we should check policy limitation
            if self.cfg.rrt.general.policy_limitation:
                if not self._get_within_policy_reach_indicator(node_rand):
                    continue

            # Get nearest node
            node_nearest = self._get_nearest_node(node_rand)

            # Get new node after simulating policy for a number of timesteps (this is the steer function)
            node_new = self._get_reach_info(node_nearest, node_rand, self.cfg.rrt.general.delta_t)

            # The new node may be 'none' in case of error, so we search another random node
            if node_new is None:
                continue

            # Debug callback
            self._debug_callback('new', locals())

            # Continue only if there is no collision until node new is reached
            if not node_new.info.done_info['collision']:

                # Get nodes near 'node_new'
                list_nodes_near = self._get_near_nodes(node_new)
                
                # Get the node with minimum cost from root
                node_min, node_with_info = self._get_min_cost_node(list_nodes_near, node_new)

                # Any error on finding 'node_min' we continue to the next iteration
                if node_min is None or node_with_info is None:
                    if not self.cfg.rrt.general.rewire:
                        node_min = node_nearest
                        node_with_info = node_new
                    else:
                        continue

                # Add node to graph
                self._set_add_node(node_min, node_new, node_with_info)

                # Debug callback
                self._debug_callback('add', locals())

                # Rewire the tree
                if self.cfg.rrt.general.rewire:
                    self._set_rewire(list_nodes_near, node_new)

            # Debug callback
            self._debug_callback('update', locals())
            
            # Save
            self.save_stats(None)
            self.save_checkpoint()

            # Log
            self.log_console()
            self.log_tensorboard()

        # Set training flag and save attrs
        self.training_over = True
        self._set_save_attributes()

    def test(self):
        """
        Test RRT.
        """

        # Get path to simulation info file
        path = Paths().exp_checkpoint('rrt', self.cfg.general.config_name_or_prefix) / 'simulation_info'

        # Return if simulation is already done
        path_part = path.parent / 'simulation_info01'
        if path_part.exists():
            self.solution_simulation_info = RRTSolutionInfo.load_from_file(path)
            return

        # Get list of waypoints
        list_nodes = self._get_solution()

        # Reset master to track stats on single episode (the first episode)
        self.env.reset_master()
        self.env.reset()

        # Place car at initial position
        self.env.car.set_pose(list_nodes[0].pose)

        # Start tracking cpu time
        cpu_time_start = time.time()

        # Navigate to next poses
        done, reward, info = False, -1, {}
        success = True
        for node in list_nodes[1:]:

            # Set target pose at next node
            self.env.target.set_pose(node.pose)

            # Control loop
            done, reward, info = False, -1, {}
            obs = self.env._get_observation()
            while not done:

                # Get action
                action = self.controller_fn(obs)

                # For RL, we use a lower frequency controller
                if self.cfg.rrt.general.controller == 'rl':

                    # Step 'n' times and update t
                    for _ in range(self.rl_action_repeat):
                        obs, reward, done, info = self.env.step(action)

                        # Break as soon as done
                        if done:
                            break

                # For MPC we step every bullet timestep
                else:
                    obs, reward, done, info = self.env.step(action)

            if info['done_info']['collision'] or info['done_info']['time']:
                success = False

        # Stop tracking cpu time
        cpu_time = time.time() - cpu_time_start

        # Set attribute and save to file
        self.solution_simulation_info = RRTSolutionInfo.new(self.env.stats.copy(deep=True), cpu_time, list_nodes,
                                                            success)
        self.solution_simulation_info.save_to_file(str(path))

        # Save new attributes
        self._set_save_attributes()

        # self._test_set_solution_simulation_info()
        # self._test_plot_solution_tree()
        # self._test_plot_solution_top_view()
        # self._test_get_solution_time()
        # self._test_get_solution_length()
        # self._test_plot_ordinal_accuracy_unique_min()
        # self._test_plot_ordinal_accuracy_top_x_unique_min()
        # self._test_plot_ordinal_accuracy_within_distance()

    def save_checkpoint(self):
        """
        Save attributes and models.
        """

        # # Saving the tree cost too much time, do it only every 50 iterations
        if len(self.list_nodes) % 50 == 0:
            self._set_save_attributes()

    def save_stats(self, epoch_info: RRTEpochInfo | None):
        """
        Save stats.
        """

        # Saving the tree cost too much time, do it only every 50 iterations
        if len(self.list_nodes) % 50 == 0:

            # Get tree
            l_nodes = copy.deepcopy(self.list_nodes)
            self.stats.l_nodes.append(l_nodes)

            # Get wall time
            self.stats.l_wall_times.append(time.time())

            # Save
            self.stats.save_to_file(self.path_stats)

    def log_console(self):
        """
        Log to console
        """

        # Log losses
        self.console.print(
            f"\n[blue bold underline]Epoch:[/] [blue]{len(self.list_nodes)}[/]\t"
        )

    def log_tensorboard(self):
        """
        Log to tensorboard
        """

        # Step is the number of nodes
        step = len(self.list_nodes)

        # Create map, add list of nodes and add figure
        mmap = Map(self.cfg)
        mmap.set_add_nodes(self.list_nodes)
        mmap.set_add_nodes(self.node_final)
        fig, _ = mmap.get_plot_tree()
        self.tb.add_figure("tree", fig, step)
        plt.close(fig)

        # Log solution found indicator (time to solve)
        self.tb.add_scalar("found_solution", int(self.found_solution), step)

        # If solution is found we track it's cost
        if self.found_solution:
            self.tb.add_scalar("solution_cost_length", self._get_solution_cost_length(), step)
            self.tb.add_scalar("solution_cost_time", self._get_solution_cost_time(), step)

    def _get_action_mpc(self, obs: dict):
        """
        Get action from pure pursuit controller.
        :param obs: current env observation (contains state and desired goal).
        :return: controller action.
        """

        return self.mpcrs.get_action(obs)

    def _get_action_rl(self, obs: dict):
        """
        Get action from agent's policy.
        :param obs: current env observation (contains state and desired goal).
        :return: controller action.
        """

        # Transform target
        new_target_pose = transform_to_origin(obs)
        obs['desired_goal'] = pose3_to_state4(new_target_pose)

        # Transform car 
        obs['achieved_goal'] = np.array([0, 0, 0, 1])
        obs['observation'][:4] = [0, 0, 0, 1]

        return self.rl.get_action(obs)

    def _get_distance_ordinal(self, node_from: Node | list[Node, ...], node_to: Node) -> Vector | float:
        """
        Get distance according to distance prediction ordinal model and maybe simulate some of them.
        :param node_from: initial node(s).
        :param node_to: final node.
        :return: distance value.
        """

        # Handle single instance case (default is parallel)
        if isinstance(node_from, Node):
            node_from = [node_from]

        # Hold input to the model (state and goal)
        n_rows = len(node_from)
        arr_goals = np.zeros((n_rows, 4))

        # Assign [xi yi si ci v phi xf yf sf cf]
        for i in range(n_rows):
            pose = transform_to_origin(None, node_from[i], node_to)
            arr_goals[i, :] = pose3_to_state4(pose)

        # Predict
        arr_goals = self.ordinal_scaler.transform(arr_goals)
        distance_pred = corn_label_from_logits(self.ordinal.predict(arr_goals)).detach().cpu().numpy().flatten()

        # Multiply by policy action repeat
        distance_pred *= self.rl_action_repeat
        
        # Check if we should simulate top-n %
        top_n_percent = int(self.cfg.rrt.general.distance_fn[-2:]) / 100
        top_n = round(top_n_percent * len(node_from))
        if top_n > 0 or self.cfg.rrt.general.distance_fn_mode == 'difference':

            # Sort ascending and get real distance for first 'n'
            distances_ascending_idxs = np.argsort(distance_pred)

            # Check the mode to select top-n nodes
            if self.cfg.rrt.general.distance_fn_mode == 'difference':
                min_dist = distance_pred[distances_ascending_idxs[0]]
                stop_at = 1
                for i, idx in enumerate(distances_ascending_idxs[1:]):
                    if np.abs(min_dist - distance_pred[idx]) > 10 * self.rl_action_repeat:
                        stop_at = i
                        break
            elif self.cfg.rrt.general.distance_fn_mode == 'difference':
                stop_at = top_n
            else:
                raise NotImplementedError

            # Sort ascending and get real distance for first 'n'
            for idx in distances_ascending_idxs[:stop_at]:

                # Check real distance
                node_new = self._get_reach_info(node_from[idx], node_to, 2000)

                # Update prediction with true value
                distance_pred[idx] = node_new.info.timesteps

        # return single number for single node distance calculation
        if distance_pred.size == 1:
            return distance_pred.item()
        else:
            return distance_pred.flatten()

    def _get_distance_rs(self, node_from: Node | list[Node, ...], node_to: Node) -> Vector | float:
        """
        Get distance according to Reed-Shepps path's length.
        :param node_from: initial node.
        :param node_to: final node.
        :return: distance value.
        """

        # Handle single distance case
        if isinstance(node_from, Node):
            node_from = [node_from]

        # Calculate distances
        distances = []
        for node in node_from:
            distances.append(self.mpcrs.get_distance(node, node_to))

        # return single number for single node distance calculation
        if len(distances) == 1:
            return distances[0]
        else:
            return np.array(distances)

    def _get_distance_sim(self, node_from, node_to):
        """
        Get distance according to simulation of agent's policy.
        :param node_from: initial node
        :param node_to: final node
        :return: distance value
        """

        # Handle single distance case
        if isinstance(node_from, Node):
            node_from = [node_from]

        # Get distance of all nodes
        distances = []
        for node in node_from:
            node_new = self._get_reach_info(node, node_to, 2000)
            distances.append(node_new.info.timesteps)

        # return single number for single node distance calculation
        if len(distances) == 1:
            return distances[0]
        else:
            return np.array(distances)

    def _get_min_cost_node(self, list_nodes_near, node_new) -> tuple[Node, Node]:
        """
        Get the node (from the list of near nodes) with minimum cost from root
        :param list_nodes_near: list nodes near 'node_new'
        :param node_new: the new node obtained after the simulation
        :return: the node with minimum cost
        """

        # Loop all nodes and track costs from root
        cost_min = np.inf
        node_min, node_final, node_with_info = None, None, None
        costs = []
        for node_near in list_nodes_near:
            
            # We simulate without time limit (we choose double the number of timesteps for near nodes)
            node_final = self._get_reach_info(node_near, node_new, 2000)

            # If the node could be reached and the cost is smaller
            if not node_final.info.done_info['collision'] and node_final.info.done_info['success']:
                cost = node_near.cost_from_root + node_final.info.cost
                costs.append(cost)
                if cost < cost_min:
                    cost_min = cost
                    node_min = node_near
                    node_with_info = node_final

            # Debug callback
            self._debug_callback('min', locals())

        return node_min, node_with_info

    def _get_near_nodes(self, node_new: Node) -> list[Node, ...]:
        """
        Get near nodes according to distance model and config cutoff value.
        :param node_new: near nodes in respect to this node
        :return: ordered list of near nodes (the nearest first)
        """

        # Get distance of all nodes to 'node_new'
        distances = self.distance_fn(self.list_nodes, node_new)

        # If there is only one distance we need wrap in a list for the next lines
        distances = np.array([distances]) if np.isscalar(distances) else distances

        # Get list of near nodes and near distances
        near_idxs = np.argwhere(np.array(distances) < self.cutoff).flatten()
        list_nodes_near = np.array(self.list_nodes)[near_idxs]
        distances_near = np.array(distances)[near_idxs]

        # Now we sort the distances, get the indexes and reorder the list of near nodes
        idxs_sorted_near_distances = np.argsort(distances_near)
        list_nodes_near_sorted = np.array(list_nodes_near)[idxs_sorted_near_distances]

        # Debug callback
        self._debug_callback('near', locals())

        return list_nodes_near_sorted

    def _get_nearest_node(self, node_rand):
        """
        Get the node that is nearest to the random node
        :param node_rand: random sampled node
        :return: nearest node
        """

        # Get distance of all nodes to 'node_new'
        distances = self.distance_fn(self.list_nodes, node_rand)
        
        # If there is only one distance we need wrap in a list for the next lines
        distances = np.array([distances]) if np.isscalar(distances) else distances
        
        # Get the index of the node with minimum distance
        nearest_idx = np.argmin(distances)

        # Debug callback
        self._debug_callback('nearest', locals())

        return self.list_nodes[nearest_idx]

    def _get_ordinal_config_and_path(self):
        """
        Get config with the best parameters for ordinal model.
        :return: the best config for ordinal model
        """

        # Get experiment checkpoint folder and config
        other_exp_name = self.cfg.rrt.ordinal.use_model_from_experiment
        if other_exp_name is not None:
            path_ckpt = Paths().exp_checkpoint('rrt', other_exp_name)
            cfg = MasterConfig.load_from_experiment('rrt', other_exp_name).copy(deep=True)
        else:
            path_ckpt = Paths().exp_checkpoint('rrt', self.cfg.general.config_name_or_prefix)
            cfg = self.cfg.copy(deep=True)

        # Check if optuna db exists
        db_path = path_ckpt / 'param_search' / 'study.db'
        db_exists = db_path.exists()

        # If it exists we need to load config and path to model from there
        if db_exists:

            # Load study
            path = 'sqlite:///' + str(db_path)
            study = optuna.create_study(study_name='study', storage=path, load_if_exists=True)

            # Get best params and set to config
            params = study.best_params
            cfg.rrt.ordinal.batch_size = params['batch_size']
            cfg.rrt.ordinal.layers = [params[f'n_units_{i}'] for i in range(params['n_layers'])]
            cfg.rrt.ordinal.n_estimators = params['n_estimators']

            # Get path to best model
            trial_num = study.best_trial.number
            model_path = str(db_path.parent / f'model_{str(trial_num).zfill(2)}')

        # Study does not exist, so we load from checkpoint folder or train from the beginning
        else:

            # If model exists at checkpoint folder we return this path
            if self._get_ordinal_model_exists_indicator(str(path_ckpt)):
                model_path = str(path_ckpt)

            # Otherwise, if search has to be performed the path depends on the best found
            else:

                # Perform search and assign new model path
                if self.cfg.rrt.ordinal.opt:
                    cfg, trial_num = self._get_ordinal_model_best_config_and_trial_num()
                    model_path = str(db_path.parent / f'model_{str(trial_num).zfill(2)}')

                # Don't search path is checkpoint folder
                else:
                    model_path = str(path_ckpt)

        return cfg, model_path

    def _get_ordinal_model_best_config_and_trial_num(self):
        """
        Search hyperparameters for the ordinal model and return the best config and trial number.
        :return: master config for best parameters
        """

        # Create path
        path = Paths().exp('rrt', self.cfg.general.config_name_or_prefix) / 'checkpoint' / 'param_search' / 'study.db'
        if not path.parent.exists():
            path.parent.mkdir()

        # Initialize study
        db_path = 'sqlite:///' + str(path)
        study = optuna.create_study(study_name='study', storage=db_path, load_if_exists=True)

        # Optimize if needed
        n_trials = 20
        path = path.parent / f'model_{n_trials - 1}'
        if not path.exists():
            ds_train = DistanceOrdinalDataset(self.cfg, 'train', None, None)
            ds_val = DistanceOrdinalDataset(self.cfg, 'val', ds_train.scaler, ds_train.timesteps_range)
            def objective(trial): return self._objective(trial, dict(cfg=self.cfg, ds_train=ds_train, ds_val=ds_val))
            study.optimize(objective, n_trials=n_trials, show_progress_bar=True)

        # Copy and override attributes
        best_cfg = self.cfg.copy(deep=True)
        best_cfg.rrt.ordinal.batch_size = study.best_params['batch_size']
        best_cfg.rrt.ordinal.layers = [study.best_params[f'n_units_{i}'] for i in range(study.best_params['n_layers'])]
        best_cfg.rrt.ordinal.n_estimators = study.best_params['n_estimators']

        return best_cfg, study.best_trial.number

    def _get_ordinal_model_exists_indicator(self, model_path):
        """
        Check if there is a save for the best ordinal model
        :param model_path: path to model folder
        :return: boolean indicating whether the model exists
        """

        # Ignore warnings
        self._set_ignore_static_warnings()

        # Check for model existence
        if len(list(iter(Path(model_path).glob("*Ensemble*")))) > 0:
            return True
        else:
            return False

    def _get_parents_recursive(self, list_nodes: list[Node]) -> list[Node]:
        """
        Find the parents recursively.
        :return: find path going backwards.
        """

        # Get the last node
        last_pose = list_nodes[-1]

        # If last node has a parent we append it to the list and call the function again
        if last_pose.parent is not None:
            list_nodes.append(last_pose.parent)
            return self._get_parents_recursive(list_nodes)

        # If last node has no parent, it's the root node
        else:
            return list_nodes

    def _get_random_node(self) -> Node:
        """
        Get a random node: may be goal node or any random node.
        :return: chosen node.
        """

        # Get SL probability (constant or gaussian)
        if np.isscalar(self.cfg.rrt.sample.learner_prob):
            sl_prob = self.cfg.rrt.sample.learner_prob
        else:
            a, b, c = self.cfg.rrt.sample.learner_prob
            sl_prob = a*np.exp(-(len(self.list_nodes) - b)**2/(2*c**2))

        # Get goal probability
        goal_prob = self.cfg.rrt.sample.goal_prob if not self.found_solution else 0

        # Sampling weights and functions
        fns = [self._get_sample_final, self._get_sample_sl, self._get_sample_uniform]
        probs = [goal_prob, sl_prob, 1 - (goal_prob + sl_prob)]
        sample_fn = self.rng.choice(fns, p=probs)

        # Get random node
        random_node = sample_fn()

        return random_node

    def _get_reach_info(self, node_from, node_to, delta_t) -> None | Node:
        """
        Simulate trying to reach one node from others.
        :param node_from: origin node
        :param node_to: destination node
        :param delta_t: maximum simulation timesteps
        :return: timesteps to reach node, success indicator (no collision) and new node (very close to 'node_to').
        """

        # For MPC we can avoid wasting time with simulation (not 100% accurate) when checking distances and collisions
        if self.cfg.rrt.general.controller == 'mpc':

            # In this case we simulate the actions from the MPC controller trying to follow the RS path
            if self.cfg.rrt.mpc.simulate:
                node_new = self._get_reach_info_bullet(node_from, node_to, delta_t)

            # Here we assume the MPC controller follows the RS path with 100% fidelity
            else:
                node_new = self._get_reach_info_rs(node_from, node_to, delta_t)

        # For RL, we must simulate
        else:
            node_new = self._get_reach_info_bullet(node_from, node_to, delta_t)

        return node_new

    def _get_reach_info_bullet(self, node_from, node_to, delta_t) -> None | Node:
        """
        Simulate trying to reach one node from others.
        :param node_from: origin node
        :param node_to: destination node
        :param delta_t: maximum simulation timesteps
        :return: timesteps to reach node, success indicator (no collision) and new node (very close to 'node_to').
        """

        # Reset env to reset timesteps counter and reset master to track stats
        self.env.reset_master()
        self.env.reset()

        # Place car and target at appropriate poses
        self.env.car.set_pose(node_from.pose)
        self.env.target.set_pose(node_to.pose)

        # Before episode beginning set flags and get initial observation
        done, info, reward = False, False, False
        obs = self.env._get_observation()

        # Loop until termination or delta t elapsed
        while self.env.timestep < delta_t:

            # Get controller's action
            action = self.controller_fn(obs)

            # For RL, we use a lower frequency controller
            if self.cfg.rrt.general.controller == 'rl':

                # Step 'n' times and update t
                for _ in range(self.rl_action_repeat):
                    obs, reward, done, info = self.env.step(action)

                    # Break as soon as done
                    if done:
                        break

            # Non-RL controllers
            else:

                # Step env one time and update t
                obs, reward, done, info = self.env.step(action)

            # Check why episode ended
            if done:
                break

        # If success we set the final pose to be equal to 'node_to' (if config is set to it)
        if not self.cfg.rrt.general.add_last_pose and info['done_info']['success']:
            pose = node_to.pose
        else:
            pose = self.env.car.get_pose()

        # We return a new node avoiding changes in the argument
        node_new = Node(pose, None, None, None, node_to.origin)
        node_new.set_info(NodeReachInfo.new_from_bullet(self.env.timestep, info['done_info'], self.env.stats))

        # Change the origin in case the random node is final but was not reach
        if node_new.origin == 'final' and not info['done_info']['success'] \
                and reward < self.cfg.rrt.general.goal_epsilon:
            node_new.origin = 'final_new'
        
        return node_new

    def _get_reach_info_rs(self, node_from, node_to, delta_t) -> None | Node:
        """
        Get reach info without simulation on pybullet.
        :param node_from: initial node
        :param node_to: destination node
        :param delta_t: maximum simulation timesteps
        :return: info about the trajectory
        """

        # Reset env to reset timesteps counter and reset master to track stats
        self.env.reset_master()
        self.env.reset()

        # Place car and target at appropriate poses
        self.env.car.set_pose(node_from.pose)
        self.env.target.set_pose(node_to.pose)

        # Before episode beginning get initial observation
        obs = self.env._get_observation()

        # Get rs path
        try:
            self.mpcrs._set_rs_path(obs)
        except TypeError:
            return None

        # Loop until termination or delta t elapsed
        reward = -1
        done_info = {}
        rs_idx, bullet_timesteps_elapsed, d = 0, 0, 0
        poses, rewards = [], []
        while bullet_timesteps_elapsed < delta_t:

            # Update current index
            rs_idx += 1
            bullet_timesteps_elapsed = np.round(rs_idx * self.cfg.rrt.rs.step_size / (1/240))

            # Place car at pose
            pose = [self.mpcrs.rs_xs[rs_idx], self.mpcrs.rs_ys[rs_idx], self.mpcrs.rs_yaws[rs_idx]]
            self.env.car.set_pose(pose)

            # Get info from env
            obs = self.env._get_observation()
            reward = self.env.compute_reward(obs['achieved_goal'], obs['desired_goal'])
            done, done_info = self.env._get_done_indicator(reward)

            # Save stats
            poses.append(pose)
            rewards.append(reward)

            # Break episode
            if done:
                break
        
        # If success we set the final pose to be equal to 'node_to'
        if not self.cfg.rrt.general.add_last_pose and done_info['success']:
            pose = node_to.pose
        else:
            pose = self.env.car.get_pose()
        
        # We return a new node avoiding changes in the argument (unless 'node_to' is the final node)
        node_new = Node(pose, None, None, None, node_to.origin)
        node_new.set_info(NodeReachInfo.new_from_rs(self.env.timestep, done_info, poses, rewards))

        # Change the origin in case the random node is final but was not reach (increase threshold)
        if node_new.origin == 'final' and not done_info['success'] \
                and reward < self.cfg.rrt.general.goal_epsilon:
            node_new.origin = 'final_new'
        
        # Get return info
        return node_new

    def _get_sample_final(self) -> Node:
        """
        Simple function to return the final node.
        :return: final node
        """

        return self.node_final

    def _get_sample_sl(self) -> Node:
        """
        Get sample from cvae.
        :return: node from generative model.
        """

        return self.rng.choice(self.list_sl_nodes)

    def _get_sample_uniform(self) -> Node:
        """
        Uniform sample of the goal space.
        :return: random node sample.
        """

        # Get random pose
        low = [0, 0, 0]
        high = [self.cfg.maps.narrow.size, self.cfg.maps.narrow.size, 2 * np.pi]

        # Continue to sample poses until there is no collision
        pose = None
        collided = True
        while collided:

            # Get random pose
            pose = self.rng.uniform(low, high)

            # Place car and check collision
            self.env.car.set_pose(pose)
            collided = self.env._get_collision_indicator()

        # Create node
        return Node(pose, None, None, None, 'random')

    def _get_solution(self) -> list[Node]:
        """
        Get the trajectory going from 'node_init' to 'node_final'.
        :return: list of poses to reach 'node_final' from 'node_init'.
        """

        # Only get trajectory if final node is reachable
        assert self.found_solution, "Final destination not found. Run for more iterations."

        # Start getting the parents from node final
        list_poses = self._get_parents_recursive([self.node_final])
        list_poses.reverse()

        return list_poses

    def _get_solution_cost_length(self):
        """
        Get length of the solution found.
        :return: length of the trajectory
        """

        # Get solution
        list_nodes = self._get_solution()

        # Initialize distance accumulator
        d_total = 0

        # Stats define how the node was reached, so we start at index 1 (initial pose has no stats)
        for node in list_nodes[1:]:
            d_total += node.info.length

        return d_total

    def _get_solution_cost_time(self):
        """
        Get the time of the solution found.
        :return: time of the trajectory
        """

        # Get solution
        list_nodes = self._get_solution()

        # Initialize distance accumulator
        t_total = 0

        # Stats define how the node was reached, so we start at index 1 (initial pose has no stats)
        for node in list_nodes[1:]:
            t_total += node.info.timesteps

        return t_total

    def _get_too_close_indicator(self, cost):
        """
        Check if new node is too close to the nearest node.
        :param cost: number of timesteps for the possible new node.
        :return:
        """

        self._setup_ignore_static_warnings()

        return True if cost < self.min_distance else False

    def _get_within_policy_reach_indicator(self, node_rand):
        """
        Check if policy would be able to reach this pose from any node on the tree,
        :param node_rand: random sampled node
        :return: flag indicating if policy would be able to reach node_rand from node_nearest
        """

        # Don't need this for RS curves
        if self.cfg.rrt.general.distance_fn == 'rs':
            return True

        # We use simple euclidean distance to check if the policy could probably reach the node
        for node in self.list_nodes:
            if euclidean(node.pose[:2], node_rand.pose[:2]) < 4.0:
                return True

        return False

    def _set_add_node(self, node_min, node_new, node_with_info):
        """
        Add new node.
        :param node_min: node that makes new node the closest to root
        :param node_new: new node to be added to the tree
        :param node_with_info: how the 'node_new' was reached from 'node_min'
        """

        # Set the parent / children and update the cost from parent
        node_new.parent = node_min
        node_min.children.append(node_new)

        # Update parent cost and parent reach info
        node_new.cost_from_parent = node_with_info.info.cost
        node_new.set_info(node_with_info.info)

        # Check if cost from root is none first (final pose has none cost from root yet, possibly)
        if node_min.cost_from_root is not None:
            node_new.cost_from_root = node_min.cost_from_root + node_with_info.info.cost

        # If the origin is 'final' the final node was reached with success
        if node_new.origin == 'final':
            self.node_final = node_new
            self.found_solution = True  
        
        # Also add node to the list of all nodes in the tree
        self.list_nodes.append(node_new)

    def _set_init_attributes(self):
        """
        Initialize attributes.
        """

        self.epoch = 0

    def _set_ordinal_model(self):
        """
        Initialize ordinal, scaler and timesteps range.
        """

        # Get the best ordinal config
        cfg, model_path = self._get_ordinal_config_and_path()

        # Initialize best ordinal attributes
        self._set_ordinal_model_init(cfg)

        # Load model
        if self._get_ordinal_model_exists_indicator(model_path):
            io.load(self.ordinal, model_path)

        # Or train model
        else:
            self._set_ordinal_model_training(cfg, model_path)

    def _set_ordinal_model_init(self, cfg: MasterConfig):
        """
        Initialize ordinal and scaler attributes
        :param cfg: the best config for the best ordinal model
        """

        # Initialize dataset
        ds_train = DistanceOrdinalDataset(cfg, 'train', None, None)

        # Initialize model and set attributes
        self.ordinal = SnapshotEnsembleRegressor(Ordinal, cfg.rrt.ordinal.n_estimators, {'cfg': cfg})
        self.ordinal_scaler = ds_train.scaler
        self.ordinal_range = ds_train.timesteps_range

        # Define criterion
        n_classes = int(ds_train.timesteps_range[1] - ds_train.timesteps_range[0] + 1)
        loss = EnsembleLoss(n_classes)

        # Configure model
        self.ordinal.set_criterion(loss)
        self.ordinal.set_optimizer('Adam', lr=1e-3)
        set_logger()

    def _set_ordinal_model_training(self, cfg: MasterConfig, model_path: str):
        """
        Train ordinal model with the best parameters.
        :param cfg: the best config for ordinal model
        """

        # Number of epochs depend on the number of estimators
        epochs = 50
        while epochs % cfg.rrt.ordinal.n_estimators != 0:
            epochs += 1

        # Get data loaders
        ds_train = DistanceOrdinalDataset(self.cfg, 'train', None, None)
        ds_val = DistanceOrdinalDataset(self.cfg, 'val', ds_train.scaler, ds_train.timesteps_range)
        dl_train = DataLoader(ds_train, batch_size=cfg.rrt.ordinal.batch_size, shuffle=True)
        dl_val = DataLoader(ds_val, batch_size=cfg.rrt.ordinal.batch_size)

        # Fit
        self.ordinal.fit(dl_train, test_loader=dl_val, epochs=epochs, save_model=True, save_dir=model_path)

    def _set_rewire(self, list_nodes_near, node_new):
        """
        Rewire step of RRT*.
        :param list_nodes_near: list of nodes near node rand
        :param node_new: new node that was added
        """

        # Loop for all near nodes except the nearest
        for node_near in list_nodes_near:

            # Don't need to test for initial node or 'node_new' parent ('node_min')
            if node_near.origin == 'init' or node_near == node_new.parent:
                continue

            # Now check if using 'node_new' as a waypoint decrease the cost from root
            node_test = self._get_reach_info(node_new, node_near, 2000)

            # None is returned if RS path fails
            if node_test is None:
                continue

            # Debug callback
            self._debug_callback('rewire', locals())

            # If no collision and the node could be reached we compare the costs
            if not node_test.info.done_info['collision'] and node_test.info.done_info['success']:
                
                # Current cost is node near cost from root
                current_cost = node_near.cost_from_root
                
                # New cost is the cost from root of the new node + the travel cost of 'new' to 'near'
                new_cost = node_new.cost_from_root + node_test.info.cost

                # If the new cost is smaller, we rewire
                if new_cost < current_cost:
                    print(f'rewiring (cost decreased) from {current_cost} to {new_cost}...')
                    # Remove 'node_near' from the list of children of the current parent node
                    node_near.set_remove_me_as_child()
                    
                    # We change the parent of the near node to be the new node
                    node_near.parent = node_new
                    node_new.children.append(node_near)
                    
                    # Update the costs and info
                    node_near.cost_from_parent = node_test.info.cost
                    node_near.cost_from_root = new_cost
                    node_near.set_info(node_test.info)
                    
                    # Propagate the change in cost to all children
                    node_near.set_cost_change_propagation()

    def _set_rl_data(self):
        """
        Check if data for ordinal model is present.
        """

        # Check if data exists and RL is loaded
        data_path = Paths().data_rl_distance / self.cfg.rrt.names.rl / 'train.csv'
        if not data_path.exists():
            self.rl.get_distance_data()

    def _test_get_time_to_solve(self):
        """
        Get the time needed to find the solution.
        :return: time needed to solve
        """

        assert self.found_solution, "Solution not found yet..."

        # Read tb event file
        path = next(iter(Paths().exp_tensorboard('rrt', self.cfg.general.config_name_or_prefix).glob("*event*")))
        reader = SummaryReader(log_path=str(path), extra_columns={'wall_time'})
        df = reader.scalars.query("tag == 'found_solution'")

        return df['wall_time'].max() - df['wall_time'].min()

    def _test_get_solution_simulation_cost_length(self):
        """
        Get the time needed to reach the final pose.
        :return: time to reach final pose
        """

        # Should run simulation with controller first
        assert self.solution_simulation_info is not None, "Run test function to set solution simulation info first..."

        return self.solution_simulation_info.ep_stats.get_distance_traveled(0)

    def _test_get_solution_simulation_cost_time(self):
        """
        Get the time needed to reach the final pose.
        :return: time to reach final pose
        """

        # Should run simulation with controller first
        assert self.solution_simulation_info is not None, "Run test function to set solution simulation info first..."

        return self.solution_simulation_info.ep_stats.get_time_to_reach(0)

    def _test_plot_ordinal_accuracy_within_distance(self):
        """
        Plot ordinal model accuracy when there is a single node with minimum distance.
        """

        path = self.path_figs / 'ordAccuracy.png'
        if path.exists():
            return

        # Get test dataset
        ds_test = DistanceOrdinalDataset(self.cfg, 'test', self.ordinal_scaler, self.ordinal_range)

        # Predict distances
        y_pred_all = corn_label_from_logits(self.ordinal.predict(ds_test.x)).detach().cpu().flatten()

        # Get maximum possible value
        y_max = ds_test.y.max()

        # Data struct
        df = pd.DataFrame(columns=['n', 'x', 'success_rate'])

        # Loop x values
        xs = [6, 8, 10]
        for x in xs:

            # Loop n values
            # ns = range(x + 1, 500 + x + 2, 50)
            ns = range(100, 550, 50)
            for n in ns:

                # Check success rate for this number of nodes
                success_counter = 0
                for t in range(100):

                    # Get min value
                    min_value_lower_than_max = False
                    y_min_idx, y_min_value = None, None
                    while not min_value_lower_than_max:
                        y_min_idx = np.random.randint(0, y_pred_all.shape[0], 1)[0]
                        y_min_value = ds_test.y[y_min_idx]
                        min_value_lower_than_max = y_min_value < y_max

                    # Get list of idxs of predictions
                    idxs = [y_min_idx]
                    while len(idxs) < n:
                        idx = np.random.randint(0, y_pred_all.shape[0], 1)[0]
                        val = ds_test.y[idx]
                        if val > y_min_value:
                            idxs.append(idx)

                    # Get values at chosen indexes
                    y_true = ds_test.y[np.array(idxs)]
                    y_pred = y_pred_all[np.array(idxs)]

                    # Sorted indexes
                    idxs_true_sorted = np.array(np.argsort(y_true))
                    idxs_pred_sorted = np.array(np.argsort(y_pred))

                    # Find index where prediction is 'x' greater than minimum prediction
                    min_pred = y_pred[idxs_pred_sorted[0]]
                    stop_at = 1
                    for i, pred in enumerate(y_pred[idxs_pred_sorted]):
                        if np.abs(pred - min_pred) > x:
                            stop_at = i
                            break

                    print(stop_at)

                    # Check if top-x from y_pred matches top-1 of true
                    if np.any(idxs_pred_sorted[:stop_at] == 0):
                        success_counter += 1

                    # Check if argmin == y_min_index (zero)
                    # if np.argmin(y_pred) == 0:
                    #     success_counter += 1

                # Add to df
                row = pd.DataFrame(dict(n=[n], x=[x], success_rate=[success_counter]))
                df = df.append(row, ignore_index=True)

        # Change formatting
        df['x'] = df['x'].apply(lambda x: str(x) + '%')

        # Rename df columns before plotting
        df.rename(columns={'x': 'Simulações', 'success_rate': 'Taxa de sucesso'}, inplace=True)

        # Plot
        plt.close()
        sns.set_style('whitegrid')
        line_plot = sns.lineplot(x="n", y="Taxa de sucesso", hue="Simulações", size="Simulações",
                                 style="Simulações",
                                 sizes=(3, 1), data=df)

        # Save
        fig = line_plot.get_figure()
        fig.savefig(path, bbox_inches="tight", dpi=600)

    def _test_plot_ordinal_accuracy_top_x_unique_min(self):
        """
        Plot ordinal model accuracy when there is a single node with minimum distance.
        """

        path = self.path_figs / 'ordAccuracyTopXPercentUniqueMin.png'
        if path.exists():
            return

        # Get test dataset
        ds_test = DistanceOrdinalDataset(self.cfg, 'test', self.ordinal_scaler, self.ordinal_range)

        # Predict distances
        y_pred_all = corn_label_from_logits(self.ordinal.predict(ds_test.x)).detach().cpu().flatten()

        # Get maximum possible value
        y_max = ds_test.y.max()

        # Data struct
        df = pd.DataFrame(columns=['n', 'x', 'success_rate'])

        # Loop x values
        xs = [10, 20, 30, 40, 50]
        for x in xs:

            # Loop n values
            # ns = range(x + 1, 500 + x + 2, 50)
            ns = range(100, 550, 50)
            for n in ns:

                # Get number of x to test
                nx = round(0.01 * x * n)

                # Check success rate for this number of nodes
                success_counter = 0
                for t in range(100):

                    # Get min value
                    min_value_lower_than_max = False
                    y_min_idx, y_min_value = None, None
                    while not min_value_lower_than_max:
                        y_min_idx = np.random.randint(0, y_pred_all.shape[0], 1)[0]
                        y_min_value = ds_test.y[y_min_idx]
                        min_value_lower_than_max = y_min_value < y_max

                    # Get list of idxs of predictions
                    idxs = [y_min_idx]
                    while len(idxs) < n:
                        idx = np.random.randint(0, y_pred_all.shape[0], 1)[0]
                        val = ds_test.y[idx]
                        if val > y_min_value:
                            idxs.append(idx)

                    # Get values at chosen indexes
                    y_true = ds_test.y[np.array(idxs)]
                    y_pred = y_pred_all[np.array(idxs)]

                    # Sorted indexes
                    idxs_true_sorted = np.array(np.argsort(y_true))
                    idxs_pred_sorted = np.array(np.argsort(y_pred))

                    # Check if top-x from y_pred matches top-1 of true
                    if np.any(idxs_pred_sorted[:nx] == 0):
                        success_counter += 1

                    # Check if argmin == y_min_index (zero)
                    # if np.argmin(y_pred) == 0:
                    #     success_counter += 1

                # Add to df
                row = pd.DataFrame(dict(n=[n], x=[x], success_rate=[success_counter]))
                df = df.append(row, ignore_index=True)

        # Change formatting
        df['x'] = df['x'].apply(lambda x: str(x) + '%')

        # Rename df columns before plotting
        df.rename(columns={'x': 'Simulações', 'success_rate': 'Taxa de sucesso'}, inplace=True)

        # Plot
        plt.close()
        sns.set_style('whitegrid')
        line_plot = sns.lineplot(x="n", y="Taxa de sucesso", hue="Simulações", size="Simulações", style="Simulações",
                                 sizes=(3, 1), data=df)

        # Save
        fig = line_plot.get_figure()
        fig.savefig(path, bbox_inches="tight", dpi=600)

    def _test_plot_ordinal_accuracy_unique_min(self):
        """
        Plot ordinal model top-x% accuracy when there is a single node with minimum distance.
        :return:
        """

        path = self.path_figs / 'ordAccuracyUniqueMin.png'
        if path.exists():
            return

        # Get test dataset
        ds_test = DistanceOrdinalDataset(self.cfg, 'test', self.ordinal_scaler, self.ordinal_range)

        # Predict distances
        y_pred_all = corn_label_from_logits(self.ordinal.predict(ds_test.x)).detach().cpu().flatten()

        # Get maximum possible value
        y_max = ds_test.y.max()

        # Data struct
        df = pd.DataFrame(columns=['n', 'x', 'success_rate'])

        ns = range(10, 510, 10)
        for n in ns:

            # Check success rate for this number of nodes
            success_counter = 0
            for t in range(100):

                # Get min value
                min_value_lower_than_max = False
                y_min_idx, y_min_value = None, None
                while not min_value_lower_than_max:
                    y_min_idx = np.random.randint(0, y_pred_all.shape[0], 1)[0]
                    y_min_value = ds_test.y[y_min_idx]
                    min_value_lower_than_max = y_min_value < y_max

                # Get list of idxs of predictions
                idxs = [y_min_idx]
                while len(idxs) < n:
                    idx = np.random.randint(0, y_pred_all.shape[0], 1)[0]
                    val = ds_test.y[idx]
                    if val > y_min_value:
                        idxs.append(idx)

                # Get values at chosen indexes
                y_pred = y_pred_all[np.array(idxs)]

                # Sorted indexes
                idxs_pred_sorted = np.array(np.argsort(y_pred))

                # Check if top-x from y_pred matches top-1 of true
                if idxs_pred_sorted[0] == 0:
                    success_counter += 1

            # Add to df
            row = pd.DataFrame(dict(n=[n], success_rate=[success_counter]))
            df = df.append(row, ignore_index=True)

        # Rename df columns before plotting
        df.rename(columns={'success_rate': 'Taxa de sucesso'}, inplace=True)

        # Plot
        plt.close()
        sns.set_style('whitegrid')
        line_plot = sns.lineplot(x="n", y="Taxa de sucesso", data=df)

        # Save
        fig = line_plot.get_figure()
        fig.savefig(path, bbox_inches="tight", dpi=600)

    def _test_plot_solution_top_view(self):
        """
        Plot top view of solution trajectory with simulation.
        """

        # Return if figure exists
        path = self.path_figs / 'rrtSolutionTopView.png'
        if path.exists():
            return

        # Should run simulation with controller first
        assert self.solution_simulation_info is not None, "Run test function to set solution simulation info first..."

        # Place ghosts 
        self.env._set_place_trajectory_ghosts(0, self.solution_simulation_info.ep_stats)

        # Take screenshot
        img = self.env._get_image_current_top_view()

        # Plot 
        fig, ax = plt.subplots()
        ax.imshow(img, origin='lower', extent=self.cfg.env.test.narrow.img_extent)
        ax.set_xlabel('x [m]')
        ax.set_ylabel('y [m]')
        fig.savefig(path, dpi=600, bbox_inches='tight')

    def _test_plot_solution_tree(self):
        """
        Plot RRT tree with a different color for the solution.
        """

        # Return if figure exists
        path = self.path_figs / 'rrtSolutionTree.png'
        if path.exists():
            return

        # Plot tree normally
        mmap = Map(self.cfg)
        mmap.set_add_nodes(self.list_nodes)
        mmap.set_add_nodes(self.node_final)
        fig, ax = mmap.get_plot_tree()

        # Now we override branches for the solution (different color and no transparency)
        mmap._set_plot_branches_for_list(ax, self._get_solution())

        # Save
        ax.set_xlabel('x [m]')
        ax.set_ylabel('y [m]')
        fig.savefig(path, dpi=600, bbox_inches='tight')

    def _setup(self, *args, **kwargs):
        """
        Set up.
        """

        # Set up base algorithm settings
        self._setup_paths()
        self._setup_folders()
        self._setup_save_config_to_file()
        self._setup_attrs_to_save()
        self._setup_init_base_attrs()
        self._setup_init_model()
        self._setup_checkpoint()
        self._setup_stats()
        self._setup_console()
        self._setup_tensorboard()

        # Set up rrt settings
        self._setup_seed()
        self._setup_cost_type()
        self._setup_conditions()
        self._setup_nodes_init_final()
        self._setup_obstacles()
        self._setup_list_nodes()
        self._setup_env()
        self._setup_sl()
        self._setup_controller_distance_fns()
        self._setup_controller_model()
        self._setup_distance_model()
        self._setup_debug()

    def _setup_attrs_to_save(self):
        """
        Set up attributes to save on checkpoint.
        """

        self.save_attrs = ['epoch', 'model_curr_loss', 'model_best_loss', 'training_over', 'found_solution',
                           'list_nodes', 'list_sl_nodes', 'node_final', 'node_init', 'narrow1_pos', 'narrow2_pos',
                           'ordinal_scaler', 'ordinal_range', 'cutoff', 'min_distance']

    def _setup_checkpoint(self):
        """
        Set up checkpoint.
        """

        # Check if there is a checkpoint
        self.load_checkpoint = self._get_checkpoint_exists_indicator()

        # Load checkpoint
        if self.load_checkpoint:

            # Load attributes
            self._set_load_attributes()

            # Load test checkpoint
            path = Paths().exp_checkpoint('rrt', self.cfg.general.config_name_or_prefix) / 'simulation_info'
            path_part = path.parent / 'simulation_info01'
            if path_part.exists():
                self.solution_simulation_info = RRTSolutionInfo.load_from_file(path)

    def _setup_conditions(self):
        """
        Setup initial (root) node and final (goal) nodes.
        """

        # Load config from experiment name
        cfg = MasterConfig.load_from_experiment('sl', self.cfg.rrt.names.sl)
        
        # If condition number is int we get from validation set
        cond = self.cfg.rrt.general.condition
        if isinstance(cond, int):
            obstacles, state_i, state_f = SL(cfg).get_conditions(cond)
            self.cfg.maps.narrow.narrow1_pos = obstacles[0:2]
            self.cfg.maps.narrow.narrow2_pos = obstacles[2:4]
            self.cfg.env.car.pose = state4_to_pose3(state_i)
            self.cfg.env.target.pose = state4_to_pose3(state_f)

        # Or generate random values
        elif isinstance(cond, str):
            if cond == 'random':
                self.cfg.maps.narrow.narrow1_pos = np.round(self.rng.uniform([3, 2], [4, 7]), 1)
                self.cfg.maps.narrow.narrow2_pos = np.round(self.rng.uniform([6, 2], [7, 7]), 1)
                self.cfg.env.car.pose = np.round(self.rng.uniform([0.8, 2.0, 0], [1.2, 7.0, 2*np.pi]), 1)
                self.cfg.env.target.pose = np.round(self.rng.uniform([8.8, 2.0, 0], [9.2, 9.0, 2*np.pi]), 1)
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    def _setup_controller_distance_fns(self):
        """
        Set up controller and distance functions according to config.
        """

        # Set controller function
        match self.cfg.rrt.general.controller:
            case "rl":
                self.controller_fn = self._get_action_rl
            case "mpc":
                self.controller_fn = self._get_action_mpc
            case _:
                raise NotImplementedError

        # Set distance function between goals functions
        match self.cfg.rrt.general.distance_fn:
            case _ if "rl-ordinal" in self.cfg.rrt.general.distance_fn:
                self.distance_fn = self._get_distance_ordinal
            case "rl-sim":
                self.distance_fn = self._get_distance_sim
            case "rs":
                self.distance_fn = self._get_distance_rs
            case _:
                raise NotImplementedError

    def _setup_controller_model(self):
        """
        Set up agent for RL control.
        """

        # Only RL controller has a model
        if self.cfg.rrt.general.controller != "rl":
            return

        # Load RL
        rl_cfg = MasterConfig.load_from_experiment('rl', self.cfg.rrt.names.rl)
        self.rl = RL(rl_cfg)

        # Set action repeat
        self.rl_action_repeat = int(1 / (rl_cfg.env.car.f_val * (1 / 240)))

    def _setup_cost_type(self):
        """
        Cutoff is different depending on the distance function.
        """

        # Don't need if loaded checkpoint
        if self.load_checkpoint:
            return
        
        match self.cfg.rrt.general.distance_fn:
            case 'rs':
                self.cutoff = self.cfg.rrt.near.cutoff_length
                self.min_distance = self.cfg.rrt.nearest.min_length
            case _:
                self.cutoff = self.cfg.rrt.near.cutoff_timesteps
                self.min_distance = self.cfg.rrt.nearest.min_n_timesteps

    def _setup_debug(self):
        """
        Debug algorithm with graphs.
        """

        if self.cfg.rrt.general.debug:
            
            # Create map with present nodes
            self.dbg_map = Map(self.cfg)
            self.dbg_map.set_add_nodes(self.list_nodes)
            self.dbg_map.set_add_nodes(self.node_final)
            
            # Initialize figure
            self.dbg_fig, self.dbg_ax = self.dbg_map.get_plot_lines()
            
            # Set legend
            # self.dbg_ax.scatter(-1, -1, color='green', label='initial node')
            # self.dbg_ax.scatter(-1, -1, color='red', label='final node')
            # self.dbg_ax.scatter(-1, -1, color='lime', label='new node')
            # self.dbg_ax.scatter(-1, -1, color='purple', label='min cost node')
            # self.dbg_ax.scatter(-1, -1, color='aqua', label='nearest node')
            # self.dbg_ax.scatter(-1, -1, color='blue', label='random node')
            # self.dbg_ax.scatter(-1, -1, color='chocolate', label='near node')
            # self.dbg_ax.legend(bbox_to_anchor=(1,1), loc="upper left")
            
            # Show
            self.dbg_fig.show()
        
        elif not self.cfg.general.is_train:
            return

        else:
            
            # Non-interactive backend
            mpl.use('agg')

    def _setup_distance_model(self):
        """
        Set up distance prediction model. This is needed for ordinal model.
        """

        # For Redd-Shepps we initialize the class and for RL we train ordinal model
        match self.cfg.rrt.general.distance_fn:
            case 'rs':
                self.mpcrs = MPCRS(self.cfg)
            case _ if 'rl-ordinal' in self.cfg.rrt.general.distance_fn:
                self._set_rl_data()
                self._set_ordinal_model()

    def _setup_env(self):
        """
        Set up PyBullet simulation.
        """

        self.cfg.env.general.gui = True if not self.cfg.general.is_train else False
        self.cfg.env.reward.epsilon *= self.cfg.rrt.general.epsilon_multiplier
        self.env = CarNavigationBulletEnv(self.cfg)

    def _setup_ignore_static_warnings(self):
        """
        Ignore static method warnings.
        """

        pass

    def _setup_init_base_attrs(self):
        """
        Initialize base attributes (epoch, model losses, etc)
        """

        # Base attributes
        self.epoch = 0
        self.model_curr_loss = np.inf
        self.model_best_loss = np.inf
        self.training_over = False
        
        # RRT attributes
        self.found_solution = False

    def _setup_list_nodes(self):
        """
        Set up list of nodes in the tree.
        """

        # Don't need if loaded checkpoint
        if self.load_checkpoint:
            return

        # Initialize graph structure
        self.list_nodes = [self.node_init]

    def _setup_nodes_init_final(self):
        """
        Set up initial and final nodes.
        """

        if self.load_checkpoint:
            return
        
        self.node_init = Node(self.cfg.env.car.pose, None, 0, 0, 'init')
        self.node_final = Node(self.cfg.env.target.pose, None, None, None, 'final')

    def _setup_obstacles(self):
        """
        Save obstacles to attribute.
        """

        if self.load_checkpoint:
            return

        self.narrow1_pos = self.cfg.maps.narrow.narrow1_pos
        self.narrow2_pos = self.cfg.maps.narrow.narrow2_pos

    def _setup_seed(self):
        """
        Set up seed for random processes. Must be called before setup SL because it is used to generate states.
        """

        if self.cfg.rrt.general.seed is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = np.random.default_rng(self.cfg.rrt.general.seed)

    def _setup_sl(self):
        """
        Set up cvae.
        """

        # Don't need if loading checkpoint
        if self.load_checkpoint:
            return

        # Load algorithm from name
        sl_cfg = MasterConfig.load_from_experiment('sl', self.cfg.rrt.names.sl)
        sl = SL(sl_cfg)

        # Initialize conditions vector
        y = []

        # Add obstacles
        x_o_1, y_o_1 = self.cfg.maps.narrow.narrow1_pos
        x_o_2, y_o_2 = self.cfg.maps.narrow.narrow2_pos
        y.extend([x_o_1, y_o_1, x_o_2, y_o_2])

        # Add initial and final pose
        y.extend(pose3_to_state4_sl(self.cfg.env.car.pose))
        y.extend(pose3_to_state4_sl(self.cfg.env.target.pose))

        # Generate states with CVAE model
        states = sl.get_samples(np.array(y), 2000, self.rng)
        states = state4_sl_to_state4_rl(states)

        # Generate list of nodes from these states
        self.list_sl_nodes = []
        for st in states:
            self.list_sl_nodes.append(Node(state4_to_pose3(st), None, None, None, "sl"))

    def _setup_stats(self):
        """
        Set up stats object.
        """

        if self.load_checkpoint:
            self.stats = RRTStats.load_from_file(self.path_stats)
        else:
            self.stats = RRTStats.new(self.cfg)
            self.stats.save_to_file(self.path_stats)

    @staticmethod
    def _objective(trial: optuna.Trial, kwargs: dict):
        """
        Objective for optuna.
        :param trial: trial argument
        :return: objective evaluation
        """

        # Get from config
        batch_size = kwargs['cfg'].rrt.ordinal.opt_batch_size
        n_layers = kwargs['cfg'].rrt.ordinal.opt_n_layers
        n_units = kwargs['cfg'].rrt.ordinal.opt_n_units
        n_estimators = kwargs['cfg'].rrt.ordinal.opt_n_estimators

        # Suggest architecture
        n_layers = trial.suggest_int('n_layers', n_layers[0], n_layers[1])
        layers = []
        for i in range(n_layers):
            layers.append(trial.suggest_int(f'n_units_{i}', n_units[0], n_units[1], step=n_units[2]))

        # Suggest batch size
        batch_size = trial.suggest_int('batch_size', batch_size[0], batch_size[1], step=batch_size[2])

        # Suggest number of estimators
        n_estimators = trial.suggest_int('n_estimators', n_estimators[0], n_estimators[1], step=n_estimators[2])

        # Get data loaders
        dl_train = DataLoader(kwargs['ds_train'], batch_size=batch_size, shuffle=True)
        dl_val = DataLoader(kwargs['ds_val'], batch_size=batch_size)

        # Number of epochs depend on the number of estimators
        epochs = 50
        while epochs % n_estimators != 0:
            epochs += 1

        # Define criterion
        n_classes = int(kwargs['ds_train'].timesteps_range[1] - kwargs['ds_train'].timesteps_range[0] + 1)
        loss = EnsembleLoss(n_classes)

        # Initialize model
        model = SnapshotEnsembleRegressor(estimator=Ordinal, n_estimators=n_estimators, cuda=True,
                                          estimator_args={'cfg': kwargs['cfg']})

        # Configure model
        model.set_criterion(loss)
        model.set_optimizer('Adam', lr=1e-3)
        set_logger()

        # Create unique name based on the current trial
        trial_num = 0
        path_base = Paths().exp('rrt', kwargs['cfg'].general.config_name_or_prefix) / 'checkpoint' / 'param_search'
        path_base = str(path_base / 'model_')
        path_full = path_base + str(trial_num).zfill(2)
        while Path(path_full).exists():
            trial_num += 1
            path_full = path_base + str(trial_num).zfill(2)

        # Fit model
        model.fit(train_loader=dl_train, epochs=epochs, test_loader=dl_val, save_dir=path_full)

        # Get validation error
        logits = model.predict(kwargs['ds_val'].x)
        y_pred = corn_label_from_logits(logits)
        mae = np.mean(np.abs(kwargs['ds_val'].y - y_pred.detach().numpy().flatten()))

        return mae

    def _debug_callback(self, mode: str, loc):
        """
        Debug callback.
        """

        # Check if it is debugging
        if self.cfg.rrt.general.debug:

            # Clear previous phase
            self.dbg_map.set_clear_temp_objs()

            # Each debugging phase is a case
            match mode:

                # Plot random node, distances from that node and nearest node
                case 'nearest':

                    # Get locals
                    node_rand = loc['node_rand']
                    distances = loc['distances']
                    nearest_idx = loc['nearest_idx']

                    # Add random node and nearest node
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_rand, 'dodgerblue')
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, self.list_nodes[nearest_idx], 'aqua')

                    # Add distances
                    for i, n in enumerate(self.list_nodes):
                        self.dbg_map.set_add_temp_text(self.dbg_ax, n.pose, f'{distances[i]:.2f}', 'black')

                # Plot new node and trajectory from nearest to this new node
                case 'new':

                    # Get locals
                    node_rand = loc['node_rand']
                    node_nearest = loc['node_nearest']
                    node_new = loc['node_new']

                    # Add random node, nearest node and new node
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_rand, 'dodgerblue')
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_nearest, 'aqua')
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_new, 'lime')

                    # Add trajectory from nearest to new
                    self.dbg_map.set_add_temp_line(self.dbg_ax, node_new, 'lime')

                # Plot new node and nodes near (with distance)
                case 'near':
                    
                    # Get locals
                    node_new = loc['node_new']
                    list_nodes_near = loc['list_nodes_near']
                    distances_near = loc['distances_near']
                    
                    # Add new node
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_new, 'lime')
                    
                    # Add near nodes
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, list_nodes_near, 'brown')

                    # Add distances
                    for i, n in enumerate(list_nodes_near):
                        self.dbg_map.set_add_temp_text(self.dbg_ax, n.pose,
                                                       f'distance (new):{distances_near[i]:.2f}', color='brown')

                # Plot new node, near nodes and the cost of new node for each near node
                case 'min':

                    # Get locals
                    node_new = loc['node_new']
                    list_nodes_near = loc['list_nodes_near']
                    costs = loc['costs']

                    # Add new node
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_new, 'lime')

                    # Add near nodes
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, list_nodes_near, 'brown')

                    # Now plot the cost from root of new node for each near node
                    for n, c in zip(list_nodes_near, costs):
                        self.dbg_map.set_add_temp_text(self.dbg_ax, n.pose,
                                                       f'cost (root): {c:.2f}', 'brown')

                # Add new node to the tree
                case 'add':

                    # Get locals
                    node_min = loc['node_min']
                    node_new = loc['node_new']
                    node_with_info = loc['node_with_info']

                    # Plot min node and new node
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_min, 'orange')
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_new, 'lime')

                    # Plot line between min and new
                    self.dbg_map.set_add_temp_line(self.dbg_ax, node_with_info, color='orange')

                # Plot rewiring for each near node
                case 'rewire':

                    # Get locals
                    node_new = loc['node_new']
                    node_near = loc['node_near']
                    node_test = loc['node_test']

                    # Add new node and this node's cost from root
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_new, 'lime')
                    self.dbg_map.set_add_temp_text(self.dbg_ax, node_new.pose,
                                                   f'cost (root): {node_new.cost_from_root:.2f}', 'lime')

                    # Add near node
                    self.dbg_map.set_add_temp_arrows(self.dbg_ax, node_near, 'purple')

                    # Add cost from new and current cost
                    self.dbg_map.set_add_temp_text(self.dbg_ax, node_near.pose,
                                                   f'cost (root): {node_near.cost_from_root:.2f}', 'purple')
                    self.dbg_map.set_add_temp_text(self.dbg_ax, [node_near.pose[0], node_near.pose[1] - 0.2],
                                                   f'cost (parent): {node_test.info.cost:.2f}', 'lime')

                # Update list of nodes of the base map
                case 'update':
                    
                    # Clear axis
                    self.dbg_ax.clear()
                    
                    # Create map with current nodes
                    self.dbg_map = Map(self.cfg)
                    self.dbg_map.set_add_nodes(self.list_nodes)
                    self.dbg_map.set_add_nodes(self.node_final)
                    
                    # Draw on axis
                    self.dbg_ax = self.dbg_map.get_plot_lines(self.dbg_ax)
                    
                    # Plot costs from root
                    for node in self.list_nodes:
                        self.dbg_map.set_add_temp_text(
                            self.dbg_ax, node.pose, f'{node.cost_from_root:.2f}', 'red'
                        )
